import os
import math
import logging
from collections import defaultdict, namedtuple
from tqdm.auto import tqdm

log = logging.getLogger(__name__)

NEG_INF = float("-inf")


MutualInformation = namedtuple("score", "words length frequency score")
CohesionScore = namedtuple(
    "score", "words length frequency cohesion_forward cohesion_backward score"
)
BranchingEntropy = namedtuple(
    "score", "words length frequency leftside_entropy rightside_entropy score"
)


def cohesion_score(
    ngrams,
    total_words=None,
    min_count=30,
    expansion_method="max",
    delimiter="_",
    **kwargs,
):
    def _max(cf, cb):
        return max(cf, cb)

    def _average(cf, cb):
        return (cf + cb) / 2

    candidates = {}
    for ngram, freq in tqdm(ngrams.items()):
        n = len(ngram)
        if n <= 1 or freq < min_count:
            continue

        cohesion_forward = pow(freq / ngrams.get(tuple(ngram[:1]), 0), 1 / (n - 1))
        cohesion_backward = pow(freq / ngrams.get(tuple(ngram[-1:]), 0), 1 / (n - 1))

        if expansion_method == "max":
            score = _max(cohesion_forward, cohesion_backward)
        elif expansion_method == "average":
            score = _average(cohesion_forward, cohesion_backward)
        elif expansion_method == "backward":
            score = cohesion_backward
        else:
            score = cohesion_forward

        candidates[ngram] = CohesionScore(
            delimiter.join(ngram),
            n,
            freq,
            cohesion_forward,
            cohesion_backward,
            score,
        )

    return candidates


def branching_entropy(
    ngrams,
    total_words=None,
    min_count=10,
    delimiter="_",
    **kwargs,
):
    def entropy(dic):
        if not dic:
            return 0.0
        sum_ = sum(dic.values())
        entropy = 0
        for freq in dic.values():
            prob = freq / sum_
            entropy += prob * math.log(prob)
        return -1 * entropy

    def parse_left(extension):
        return extension[:-1]

    def parse_right(extension):
        return extension[1:]

    def sort_by_length(counter, min_count):
        sorted_by_length = defaultdict(lambda: [])
        for ngram, freq in counter.items():
            if freq < min_count or len(ngram) <= 2:
                continue
            sorted_by_length[len(ngram)].append(ngram)
        return sorted_by_length

    def get_entropy_table(parse, sorted_by_length):
        be = {}
        for n, ngram_list in tqdm(sorted_by_length.items()):
            extensions = defaultdict(lambda: [])
            for ngram in ngram_list:
                extensions[parse(ngram)].append(ngram)
            for ngram, extension_ngrams in extensions.items():
                extension_frequency = {
                    ext: ngrams.get(ext, 0) for ext in extension_ngrams if ext in ngrams
                }
                be[ngram] = entropy(extension_frequency)
        return be

    sorted_by_length = sort_by_length(ngrams, min_count)
    be_l = get_entropy_table(parse_right, sorted_by_length)
    be_r = get_entropy_table(parse_left, sorted_by_length)
    bes = {ngram: (bel, be_r.get(ngram, 0)) for ngram, bel in be_l.items()}
    for ngram, ber in be_r.items():
        if not (ngram in be_l):
            bes[ngram] = (0, ber)

    candidates = {}
    for ngram, (bel, ber) in bes.items():
        score = bel * ber
        candidates[ngram] = BranchingEntropy(
            delimiter.join(ngram),
            len(ngram),
            ngrams.get(ngram, 0),
            bel,
            ber,
            score,
        )

    return candidates


def mutual_information(
    ngrams,
    total_words=None,
    delta=0.0,
    expansion_method="average",
    normalize=False,
    delimiter="_",
    **kwargs,
):
    def _average(scores):
        return 0 if not scores else sum(scores) / len(scores)

    def _max(scores):
        return 0 if not scores else max(scores)

    def _top3_average(scores):
        return (
            0
            if not scores
            else sum(sorted(scores, reverse=True)[:3]) / min(3, len(scores))
        )

    # max_n = max([len(ngram) for ngram in ngrams.keys()])
    candidates = {}
    total_words = float(total_words)
    num_ngrams = len(ngrams)
    for ngram, ab in tqdm(ngrams.items()):
        if (len(ngram) == 1) or (ab <= delta):
            continue
        score_candidates = {}
        for i in range(1, len(ngram)):
            a = ngrams.get(tuple(ngram[:i]), 0)
            b = ngrams.get(tuple(ngram[i:]), 0)
            if (a == 0) or (b == 0):
                continue
            if normalize:
                pa = a / total_words
                pb = b / total_words
                pab = ab / total_words
                score = math.log(pab / (pa * pb)) / -math.log(pab)
            else:
                score = (ab - delta) / float(a * b) * num_ngrams
            score_candidates[i] = score

        if not score_candidates:
            continue

        if expansion_method == "max":
            score = _max(score_candidates.values())
        elif expansion_method == "top3_average":
            score = _top3_average(score_candidates.values())
        else:
            score = sum(score_candidates.values()) / len(score_candidates)

        candidates[ngram] = MutualInformation(
            delimiter.join(ngram), num_ngrams, ab, score
        )
    return candidates


def get_available_memory():
    """It returns remained memory as percentage"""
    import psutil

    mem = psutil.virtual_memory()
    return 100 * mem.available / (mem.total)


def get_process_memory():
    """It returns the memory usage of current process"""
    import psutil

    process = psutil.Process(os.getpid())
    return process.memory_info().rss / (1024 ** 3)


def prune_vocab(vocab, prune_min_count):
    """Remove all entries from the `vocab` dictionary with count smaller than `prune_min_count`.
    Modifies `vocab` in place, returns the sum of all counts that were pruned.
    Parameters
    ----------
    vocab : dict
        Input dictionary.
    prune_min_count : int
        Frequency threshold for tokens in `vocab`.
    Returns
    -------
    result : int
        Sum of all counts that were pruned.
    """
    result = 0
    old_len = len(vocab)
    for w in list(vocab):  # make a copy of dict's keys
        if vocab[w] <= prune_min_count:
            result += vocab[w]
            del vocab[w]
    log.info(
        "pruned out %i tokens with count <=%i (before %i, after %i)",
        old_len - len(vocab),
        prune_min_count,
        old_len,
        len(vocab),
    )
    return result



import logging
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from ekorpkit.pipelines.pipe import apply_pipeline


log = logging.getLogger(__name__)


def build_corpus(**args):
    cfg = args.get(eKonf.Keys.CORPUS, {}).get("builtin", None)
    # print(cfg)
    if cfg:
        DatasetBuilder(**cfg)


def build_t5(**args):
    cfg = args.get(eKonf.Keys.DATASET, {}).get("t5", None)
    # print(cfg)
    if cfg:
        DatasetBuilder(**cfg)


def build_simple(**args):
    cfg = args.get(eKonf.Keys.DATASET, {}).get("simple", None)
    # print(cfg)
    if cfg:
        DatasetBuilder(**cfg)


class DatasetBuilder:
    def __init__(self, **args) -> None:
        args = eKonf.to_config(args)
        self.args = args
        self.name = args.name
        self.data_dir = args.data_dir
        self.data_filetype = args.get("filetype", ".parquet")
        self.column_info = self.args.column_info
        self.verbose = self.args.get("verbose", False)
        self.auto = self.args.auto

        self._io_ = args.io
        self.fetch_dir = self._io_.get("data_dir", None)
        self.fetch_sources = self._io_.get("data_sources", None)
        if isinstance(self.fetch_sources, str):
            self.fetch_sources = [self.fetch_sources]
        if isinstance(self.fetch_sources, list):
            self.fetch_sources = {"train": self.fetch_sources}
        self.num_workers = self._io_.get("num_workers", None)
        self.force = self._io_.force

        self.fetcher = self._io_.get("fetcher", None)
        self.loader = self._io_.get("loader", None)

        self._info_ = self.args.get("info", None)
        self.summary_info = None

        self._pipeline_ = self.args.get("pipeline", {})
        self.transform_pipeline = self._pipeline_.get("_transform_", [])
        self.process_pipeline = self._pipeline_.get("_preprocess_", [])
        if self.transform_pipeline is None:
            self.transform_pipeline = []
        if self.process_pipeline is None:
            self.process_pipeline = []

        if self.auto.load:
            self.build()

    def build(self):
        if self.fetcher:
            if eKonf.is_instantiatable(self.fetcher):
                eKonf.instantiate(self.fetcher)
            _pipeline_ = self.fetcher.get("pipeline", None)
            if _pipeline_:
                eKonf.instantiate(_pipeline_)

        if self._info_:
            self.summary_info = eKonf.instantiate(self._info_)
        if self.summary_info:
            self.summary_info.load(self.args)

        for split_name, split_data_source in self.fetch_sources.items():
            if split_data_source is None:
                continue
            self._process_split(split_name)

        if self.summary_info:
            self.summary_info.save()

        log.info(
            f"\nCorpus [{self.name}] is built to [{self.data_dir}] from [{self.fetch_dir}]"
        )

    def _process_split(self, split_name):

        _data_path_ = self._io_.data.path[split_name]
        _meta_path_ = self._io_.meta.path[split_name]
        _sample_path_ = self._io_.sample.path[split_name]

        pipe = "save_metadata"
        if pipe in self._pipeline_:
            if pipe not in self.transform_pipeline:
                self.transform_pipeline.append(pipe)
            self._pipeline_[pipe].path.output = _meta_path_
            self._pipeline_[pipe].column_info = self.column_info
            self._pipeline_[pipe].split_name = split_name
        pipe = "save_samples"
        if pipe in self._pipeline_:
            if pipe not in self.process_pipeline:
                self.process_pipeline.append(pipe)
            self._pipeline_[pipe].path.output = _sample_path_

        df = None
        if not eKonf.exists(_data_path_.filepath) or self.force.build:
            with elapsed_timer(format_time=True) as elapsed:
                df = eKonf.instantiate(self.loader, split_name=split_name)
                log.info(f" >> elapsed time to load and parse data: {elapsed()}")

            if df is None:
                raise ValueError("dataframe is None")

            if self.verbose:
                print(df.head())
                print(df.shape)

            if self.transform_pipeline and len(self.transform_pipeline) > 0:
                log.info(
                    f"\nTransforming dataframe with pipeline: {self.transform_pipeline}"
                )
                df = apply_pipeline(df, self.transform_pipeline, self._pipeline_)

            # Saving data
            columns = self.column_info.data
            if columns:
                columns = list(columns.keys())
            _data_path_.columns = columns
            eKonf.save_data(df, **_data_path_)

            if self.summary_info and self.force.summarize:
                stats = {
                    "name": split_name,
                    "dataset_name": self.name,
                    "data_file": _data_path_.filename,
                }
                if eKonf.exists(_meta_path_.filepath):
                    stats["meta_file"] = _meta_path_.filename
                self.summary_info.init_stats(df=df, split_name=split_name, stats=stats)

        else:
            log.info(f"{_data_path_.filepath} already exists")
            if self.force.summarize or self.force.preprocess:
                df = eKonf.load_data(**_data_path_)

        if df is None:
            log.warning("No datasets found")
            return None

        if self.process_pipeline and len(self.process_pipeline) > 0:
            log.info(f"\nProcessing dataframe with pipeline: {self.process_pipeline}")
            df = apply_pipeline(df, self.process_pipeline, self._pipeline_)

        if self.force.summarize and self.summary_info:
            self.summary_info.calculate_stats(df, split_name)



from ...labeling.apply.dask import (  # pragma: no cover
    DaskLFApplier,
    PandasParallelLFApplier,
)


class DaskSFApplier(DaskLFApplier):  # pragma: no cover
    """SF applier for a Dask DataFrame.

    See ``snorkel.labeling.apply.dask.DaskLFApplier`` for details.
    """

    _use_recarray = True


class PandasParallelSFApplier(PandasParallelLFApplier):  # pragma: no cover
    """Parallel SF applier for a Pandas DataFrame.

    See ``snorkel.labeling.apply.dask.PandasParallelLFApplier`` for details.
    """

    _use_recarray = True



from pathlib import Path


def save_wordlist(words, filepath, sort=True, verbose=True, **kwargs):
    if sort:
        words = sorted(words)
    if verbose:
        print(
            "Save the list to the file: {}, no. of words: {}".format(
                filepath, len(words)
            )
        )
    with open(filepath, "w") as f:
        for word in words:
            f.write(word + "\n")


def load_wordlist(
    filepath,
    rewrite=False,
    max_ngram=None,
    remove_tag=False,
    remove_delimiter=False,
    remove_duplicate=True,
    sort=True,
    lowercase=False,
    verbose=True,
    **kwargs,
):
    filepath = Path(filepath)
    if filepath.is_file():
        with open(filepath) as f:
            words = [word.strip().split()[0] for word in f if len(word.strip()) > 0]
    else:
        words = []
        save_wordlist(words, filepath, verbose=verbose)

    if remove_delimiter:
        words = [word.replace(";", "") for word in words]
    if sort and remove_duplicate:
        words = sorted(set(words))
    if max_ngram:
        words = [word for word in words if len(word.split(";")) <= max_ngram]
    if verbose:
        print("Loaded the file: {}, No. of words: {}".format(filepath, len(words)))
    if rewrite:
        if sort:
            words = sorted(words)
        with open(filepath, "w") as f:
            for word in words:
                f.write(word + "\n")
        if verbose:
            print("Rewrite the file: {}, No. of words: {}".format(filepath, len(words)))

    if remove_tag:
        words = [word.split("/")[0] for word in words]
    words = [
        word.lower() if lowercase else word
        for word in words
        if not word.startswith("#")
    ]
    if remove_duplicate:
        words = list(set(words))
        if verbose:
            print("Remove duplicated words, No. of words: {}".format(len(words)))
    return words



import logging
import os
import numpy as np
import matplotlib.pyplot as plt
from .base import _configure_font
from PIL import Image, ImageDraw, ImageFont
from ekorpkit.io.file import get_filepaths


log = logging.getLogger(__name__)


def convert_image(
    img_file,
    show_filename=False,
    filename_offset=(5, 5),
    fontname=None,
    fontsize=12,
    fontcolor=None,
):
    img = Image.open(img_file)
    if show_filename:
        font = _get_imagefont(fontname, fontsize)
        draw = ImageDraw.Draw(img)
        draw.text(
            filename_offset, os.path.basename(img_file), font=font, fill=fontcolor
        )

    img = img.convert("RGB")
    img = np.asarray(img)
    return img


def _get_imagefont(fontname=None, fontsize=12):
    fontname, fontpath = _configure_font(set_font_for_matplot=False, fontname=fontname)
    if fontpath:
        font = ImageFont.truetype(fontpath, fontsize)
    else:
        font = None
    return font


def gallery(array, ncols=7):
    nindex, height, width, intensity = array.shape
    nrows = nindex // ncols
    assert nindex == nrows * ncols
    # want result.shape = (height*nrows, width*ncols, intensity)
    result = (
        array.reshape(nrows, ncols, height, width, intensity)
        .swapaxes(1, 2)
        .reshape(height * nrows, width * ncols, intensity)
    )

    return result


def collage(
    image_filepaths=None,
    filename_patterns=None,
    base_dir=None,
    output_filepath=None,
    ncols=7,
    num_images=None,
    figsize=(30, 20),
    dpi=300,
    title=None,
    title_fontsize=12,
    show_filename=False,
    filename_offset=(5, 5),
    fontname=None,
    fontsize=12,
    fontcolor="#000",
    xlabel=None,
    ylabel=None,
    xticklabels=None,
    yticklabels=None,
    xlabel_fontsize=12,
    ylabel_fontsize=12,
    **kwargs,
):
    verbose = kwargs.get("verbose", False)
    if image_filepaths is None:
        image_filepaths = sorted(get_filepaths(filename_patterns, base_dir=base_dir))
    if not image_filepaths:
        log.warning("no images found")
        return

    img_arr = []
    for filepath in image_filepaths:
        img_arr.append(
            convert_image(
                filepath, show_filename, filename_offset, fontname, fontsize, fontcolor
            )
        )
    if num_images is not None:
        num_images = min(num_images, len(img_arr))
    else:
        num_images = len(img_arr)
    ncols = min(ncols, num_images)
    nrows = num_images // ncols
    num_images = nrows * ncols

    img_arr = img_arr[:num_images]
    array = np.array(img_arr)
    result = gallery(array, ncols=ncols)

    plt.figure(figsize=figsize)
    plt.imshow(result)
    ax = plt.gca()
    plt.grid(False)
    if xlabel is None and ylabel is None:
        plt.axis("off")
    if title is not None:
        plt.title(title, fontdict={"fontsize": title_fontsize})
    if xlabel is not None:
        plt.xlabel(xlabel, fontdict={"fontsize": xlabel_fontsize})
    if ylabel is not None:
        plt.ylabel(ylabel, fontdict={"fontsize": ylabel_fontsize})
    if xticklabels is not None:
        # get ncols number of xticks from xlim
        xlim = ax.get_xlim()
        xticks = np.linspace(xlim[0], xlim[1], ncols + 1)
        xticks = xticks - (xticks[1] - xticks[0]) / 2
        xticks[0] = xlim[0]
        ax.set_xticks(xticks)
        xticklabels = [""] + xticklabels
        ax.set_xticklabels(xticklabels, fontsize=xlabel_fontsize)
    if yticklabels is not None:
        # get nrows number of yticks from ylim
        ylim = ax.get_ylim()
        yticks = np.linspace(ylim[0], ylim[1], nrows + 1)
        yticks = yticks - (yticks[1] - yticks[0]) / 2
        yticks[0] = ylim[0]
        ax.set_yticks(yticks)
        yticklabels = [""] + yticklabels
        ax.set_yticklabels(yticklabels, fontsize=ylabel_fontsize)

    plt.tight_layout()
    plt.show()
    if output_filepath is not None:
        if base_dir is not None:
            output_filepath = os.path.join(base_dir, output_filepath)
        ax.figure.savefig(output_filepath, dpi=dpi)

    return output_filepath



"""
AWS S3.
"""

import io
from typing import Optional, Tuple

from ..common import _split_cloud_path
from .scheme_client import SchemeClient


class S3Client(SchemeClient):
    try:
        from botocore.exceptions import EndpointConnectionError
    except ImportError:
        raise ImportError(
            "S3Client requires botocore to be installed. "
            "Install it with `pip install botocore`."
        )

    recoverable_errors = SchemeClient.recoverable_errors + (EndpointConnectionError,)
    scheme = "s3"

    def __init__(self, resource: str) -> None:
        import boto3
        import botocore

        super().__init__(resource)
        bucket_name, s3_path = S3Client.split_s3_path(resource)
        session = boto3.session.Session()
        if session.get_credentials() is None:
            # Use unsigned requests.
            s3_resource = session.resource(
                "s3", config=botocore.client.Config(signature_version=botocore.UNSIGNED)
            )
        else:
            s3_resource = session.resource("s3")
        self.s3_object = s3_resource.Object(bucket_name, s3_path)
        self._loaded = False

    def load(self):
        from botocore.exceptions import ClientError

        if not self._loaded:
            try:
                self.s3_object.load()
                self._loaded = True
            except ClientError as exc:
                if int(exc.response["Error"]["Code"]) == 404:
                    raise FileNotFoundError("file {} not found".format(self.resource))
                else:
                    raise

    def get_etag(self) -> Optional[str]:
        self.load()
        return self.s3_object.e_tag

    def get_size(self) -> Optional[int]:
        self.load()
        return self.s3_object.content_length

    def get_resource(self, temp_file: io.BufferedWriter) -> None:
        self.load()
        self.s3_object.download_fileobj(temp_file)

    @staticmethod
    def split_s3_path(url: str) -> Tuple[str, str]:
        return _split_cloud_path(url, "s3")



import logging
import os
from abc import ABCMeta, abstractmethod
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class BaseFetcher:
    __metaclass__ = ABCMeta

    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        self.name = self.args.name
        self.auto = self.args.auto
        self.force = self.args.force
        self.verbose = self.args.get("verbose", True)

        self._path = self.args.path
        self.output_dir = self._path.output.base_dir
        self.output_file = self._path.output.filepath
        self._data = None

    def fetch(self):
        if not self._exsits() or self.force.download:
            self._fetch()
        else:
            log.info(f"{self.output_file} already exists. skipping..")
            self._load()

    def _load(self):
        self._data = eKonf.load_data(self.output_file)

    def _exsits(self):
        return os.path.exists(self.output_file)

    def _fetch(self):
        raise NotImplementedError

    @property
    def data(self):
        return self._data



import os
import sys
import re
import requests
from bs4 import BeautifulSoup
from datetime import datetime
import pdfplumber

# Import parent class
from .base import FOMC


class MeetingScript(FOMC):
    """
    A convenient class for extracting meeting scripts from the FOMC website.
    FOMC publishes the meeting scripts after 5 years, so this cannot be used for the prediction of the monetary policy in real-time.
    """

    def __init__(self, content_type, **args):
        super().__init__(content_type, **args)
        left_pct = 0.05  # % Distance of left side of character from left side of page.
        top_pct = 0.10  # % Distance of top of character from top of page.
        right_pct = (
            0.95  # % Distance of right side of character from left side of page.
        )
        bottom_pct = 0.88  #% Distance of bottom of the character from top of page.
        self.crop_coords = [left_pct, top_pct, right_pct, bottom_pct]

    def _get_links(self, from_year):
        """
        Override private function that sets all the links for the contents to download on FOMC website
         from from_year (=min(2015, from_year)) to the current most recent year
        """
        self.links = []
        self.titles = []
        self.speakers = []
        self.dates = []

        r = requests.get(self.calendar_url)
        soup = BeautifulSoup(r.text, "html.parser")

        # Meeting Script can be found only in the archive as it is published after five years
        if from_year > 2014:
            print("Meeting scripts are available for 2014 or older")
        if from_year <= 2014:
            for year in range(from_year, 2015):
                yearly_contents = []
                fomc_yearly_url = (
                    self.base_url
                    + "/monetarypolicy/fomchistorical"
                    + str(year)
                    + ".htm"
                )
                r_year = requests.get(fomc_yearly_url)
                soup_yearly = BeautifulSoup(r_year.text, "html.parser")
                meeting_scripts = soup_yearly.find_all(
                    "a", href=re.compile("^/monetarypolicy/files/FOMC\d{8}meeting.pdf")
                )
                for meeting_script in meeting_scripts:
                    self.links.append(meeting_script.attrs["href"])
                    self.speakers.append(
                        self._speaker_from_date(
                            self._date_from_link(meeting_script.attrs["href"])
                        )
                    )
                    self.titles.append("FOMC Meeting Transcript")
                    self.dates.append(
                        datetime.strptime(
                            self._date_from_link(meeting_script.attrs["href"]),
                            "%Y-%m-%d",
                        )
                    )
                if self.verbose:
                    print(
                        "YEAR: {} - {} meeting scripts found.".format(
                            year, len(meeting_scripts)
                        )
                    )
            print("There are total ", len(self.links), " links for ", self.content_type)

    def _add_article(self, link, index=None):
        """
        Override a private function that adds a related article for 1 link into the instance variable
        The index is the index in the article to add to.
        Due to concurrent processing, we need to make sure the articles are stored in the right order
        """
        if self.verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

        link_url = self.base_url + link
        pdf_filepath = (
            self.output_raw_dir
            + "/FOMC_MeetingScript_"
            + self._date_from_link(link)
            + ".pdf"
        )

        if not os.path.exists(pdf_filepath) or self.force_download:
            # Scripts are provided only in pdf. Save the pdf and pass the content
            res = requests.get(link_url)
            with open(pdf_filepath, "wb") as f:
                f.write(res.content)
        else:
            if self.verbose:
                print("File already exists: ", pdf_filepath)
        # Extract text from the pdf
        pdf_file_parsed = ""  # new line
        with pdfplumber.open(pdf_filepath) as pdf:
            for page in pdf.pages:
                pg_width = page.width
                pg_height = page.height
                pg_bbox = (
                    self.crop_coords[0] * float(pg_width),
                    self.crop_coords[1] * float(pg_height),
                    self.crop_coords[2] * float(pg_width),
                    self.crop_coords[3] * float(pg_height),
                )
                page_crop = page.crop(bbox=pg_bbox)
                text = page_crop.extract_text()
                pdf_file_parsed = pdf_file_parsed + "\n" + text
        paragraphs = re.sub("(\n)(\n)+", "\n", pdf_file_parsed.strip())
        paragraphs = paragraphs.split("\n")

        section = -1
        paragraph_sections = []
        for paragraph in paragraphs:
            if not re.search(
                "^(page|january|february|march|april|may|june|july|august|september|october|november|december|jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)",
                paragraph.lower(),
            ):
                if len(re.findall(r"[A-Z]", paragraph[:10])) > 5 and not re.search(
                    "(present|frb/us|abs cdo|libor|rp–ioer|lsaps|cusip|nairu|s cpi|clos, r)",
                    paragraph[:10].lower(),
                ):
                    section += 1
                    paragraph_sections.append("")
                if section >= 0:
                    paragraph_sections[section] += paragraph
        self.articles[index] = self.segment_separator.join(
            [paragraph for paragraph in paragraph_sections]
        )



import os
import pandas as pd
import logging
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from .dataset import Dataset
from .base import BaseSet


log = logging.getLogger(__name__)


class Datasets(BaseSet):

    SPLITS = eKonf.SPLITS

    def __init__(self, **args):
        super().__init__(**args)
        self.datasets = args.get("datasets", None)
        self.datasets = eKonf.to_dict(self.datasets)
        if self.datasets is None:
            self.datasets = self.name
        if isinstance(self.datasets, str):
            self.datasets = {self.datasets: None}
        elif eKonf.is_list(self.datasets):
            self.datasets = {name: None for name in self.datasets}
        if self.name is None and eKonf.is_dict(self.datasets):
            self.name = "-".join(self.datasets.keys())

        self._info = args.copy()
        self._info["name"] = self.name
        self._info["datasets"] = self.datasets

        self._method_ = self.args.get("_method_", None)
        use_name_as_subdir = args.get("use_name_as_subdir", True)

        self.load_column_info()

        self._datasets_concatenated = False

        with elapsed_timer(format_time=True) as elapsed:
            for name in self.datasets:
                log.info(f"processing {name}")
                args["name"] = name
                args["data_dir"] = self.data_dir
                args["use_name_as_subdir"] = use_name_as_subdir
                args["verbose"] = self.verbose
                if self.data_files is not None:
                    if name in self.data_files:
                        args["data_files"] = self.data_files[name]
                    elif "train" in self.data_files:
                        args["data_files"] = self.data_files
                dataset = Dataset(**args)
                self.datasets[name] = dataset
                if self.splits is None:
                    self.splits = {split: None for split in dataset.data_files}
            log.info(f">>> Elapsed time: {elapsed()} <<< ")

        eKonf.methods(self._method_, self)

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname}\n----------\n"
        for name in self.datasets.keys():
            s += f"{str(name)}\n"
        return s

    def __iter__(self):
        for dataset in self.datasets.values():
            yield dataset

    def __getitem__(self, name):
        if name not in self.datasets:
            raise KeyError(f"{name} not in datasets")
        return self.datasets[name]

    def __len__(self):
        return len(self.datasets)

    def load(self):
        for _name in self.datasets:
            self.datasets[_name].load()
        self._loaded = True

    def concat_datasets(self, append_dataset_name=True):
        self.concatenate(append_name=append_dataset_name)

    def concatenate(self, append_name=True):
        if not self._loaded:
            self.load()

        dfs = []
        for name in self.datasets:
            df = self.datasets[name][self.SPLITS.TRAIN]
            dfs.append(df)
        common_columns = self.COLUMN.common_columns(dfs)

        for split in self.splits:
            dfs = []
            for name in self.datasets:
                df = self.datasets[name][split]
                if df is None:
                    continue
                if common_columns:
                    df = df[common_columns].copy()
                if append_name:
                    df = self.COLUMN.append_dataset(df, name)
                dfs.append(df)
            if len(dfs) > 1:
                self._splits[split] = pd.concat(dfs, ignore_index=True)
        if self.verbose:
            log.info(f"concatenated {len(self.datasets)} dataset(s)")
        self._datasets_concatenated = True

    def persist(self):
        if len(self.datasets) < 2:
            log.warning(f"more than one dataset required to persist")
            return
        if not self._datasets_concatenated:
            log.warning(f"datasets not concatenated yet, calling concatenate()")
            self.concatenate()

        data_dir = os.path.join(self.data_dir, self.name)

        summary_info = None
        if self._info_cfg:
            self._info_cfg["data_dir"] = data_dir
            self._info_cfg["name"] = self.name
            self._info_cfg["info_file"] = None
            summary_info = eKonf.instantiate(self._info_cfg)
        if summary_info:
            summary_info.load(self._info)

        for split, data in self.splits.items():
            if data is None:
                continue
            data_file = f"{self.name}-{split}{self.filetype}"
            data = self.COLUMN.reset_id(data)
            eKonf.save_data(data, data_file, data_dir)
            if summary_info:
                stats = {"data_file": data_file}
                summary_info.init_stats(split_name=split, stats=stats)
                summary_info.calculate_stats(data, split)
        if summary_info:
            summary_info.save(info={"column_info": self.COLUMN.INFO})



from itertools import groupby
import logging
import pandas as pd
import numpy as np
from enum import Enum
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer


log = logging.getLogger(__name__)


class _Keys(str, Enum):
    DEFAULT = eKonf.Keys.DEFAULT.value
    FEATURE = "feature"
    FEATURES = "features"
    ARTICLE_FEATURES = "article_features"
    AGGREGATE_SCORES = "aggregate_scores"
    LEXICON_FEATURES = "lexicon_features"
    POLARITY = "polarity"
    NUM_TOKENS = "num_tokens"
    NUM_EXAMPLES = "num_examples"
    COUNT = "count"
    CONDITIONS = "conditions"
    APPLY = "apply"
    AGGS = "aggs"
    EVALS = "evals"
    SCORE = "score"
    SCORES = "scores"
    LABELS = "labels"
    LABEL_BY = "label_by"
    AGG_METHOD = "agg_method"


class BaseSentimentAnalyser:
    """
    A base class for sentiment analysis.
    """

    EPSILON = 1e-6

    def __init__(self, **args):
        args = eKonf.to_config(args)
        self.args = args
        self._predict_ = args[eKonf.Keys.PREDICT]

        self._predict_.features = (
            self._predict_.get(_Keys.FEATURES) or _Keys.POLARITY.value
        )
        self._eval_ = args.get(eKonf.Keys.EVAL)
        self._features = args.get(_Keys.FEATURES)
        self._article_features = args.get(_Keys.ARTICLE_FEATURES)
        self._aggregate_scores = args.get(_Keys.AGGREGATE_SCORES)
        self.num_workers = args.num_workers
        self.use_batcher = args.use_batcher
        self.verbose = args.get("verbose", False)

        self._ngram = None
        self._sentence_separator = eKonf.Defaults.SENT_SEP.value

    def analyze(self, text, **kwargs):
        """
        :type text: str
        :returns: list
        """
        raise NotImplementedError

    def tokenize(self, text):
        """Tokenize text.

        :returns: list
        """
        raise NotImplementedError

    def predict(self, data, _predict_={}):
        if _predict_:
            self._predict_ = _predict_

        input_col = self._predict_[eKonf.Keys.INPUT]
        log.info(
            f"Predicting sentiments of the column [{input_col}] using {self._predict_}"
        )
        with elapsed_timer(format_time=True) as elapsed:
            predictions = eKonf.apply(
                self.predict_sentence,
                data[input_col],
                description=f"Predicting [{input_col}]",
                verbose=self.verbose,
                use_batcher=self.use_batcher,
                num_workers=self.num_workers,
            )
            pred_df = pd.DataFrame(predictions.tolist(), index=predictions.index)
            data = data.join(pred_df)
            log.info(" >> elapsed time to predict: {}".format(elapsed()))

        return data

    def predict_sentence(self, text, features="polarity", min_examples=5):
        """Get score for a list of terms.

        :returns: dict
        """
        raise NotImplementedError

    def predict_article(
        self,
        article,
        features=["polarity"],
        min_examples=2,
        **kwargs,
    ):
        """Predict sentiment for an article.

        :returns: dict
        """
        raise NotImplementedError

    def aggregate_scores(
        self,
        data,
        groupby,
        feature="polarity",
        min_examples=2,
        _method_=None,
    ):
        """Get aggreagate score for features.

        :returns: dataframe
        """
        _aggregate_scores = self._aggregate_scores
        _method_ = _method_ or feature
        _agg_method = (
            _aggregate_scores.get(_method_) or _aggregate_scores[_Keys.DEFAULT]
        )
        _num_examples = _Keys.NUM_EXAMPLES.value

        groupby = eKonf.ensure_list(groupby)
        if feature in data.columns:
            data = data.copy().dropna(subset=[feature])
        else:
            data = data.copy()

        eps = self.EPSILON
        _conditions = _agg_method.get(_Keys.CONDITIONS)
        _apply = _agg_method.get(_Keys.APPLY)
        _aggs = eKonf.to_dict(_agg_method.get(_Keys.AGGS))
        _evals = _agg_method.get(_Keys.EVALS)
        _scores = _agg_method.get(_Keys.SCORES)
        # _label_by = _agg_method.get(_Keys.LABEL_BY)
        _scores = {
            _name: _name.replace(_Keys.FEATURE.value, feature) for _name in _scores
        }
        _labels = self._article_features.get(feature, {}).get(_Keys.LABELS)

        data.rename(columns={feature: _Keys.FEATURE.value}, inplace=True)
        if self.verbose > 5:
            log.info("Evaluating %s", feature)
            print(data.head())
        if _apply is not None:
            for _name, _expr in _apply.items():
                data[_name] = data.apply(lambda x: eval(_expr), axis=1)
        if _conditions is not None:
            for _name, _expr in _conditions.items():
                data[_name] = np.where(data.eval(_expr), 1, 0)

        _aggs = {k: v for k, v in _aggs.items() if k in data.columns}
        agg_scores = data.groupby(groupby).agg(_aggs)
        agg_scores.columns = agg_scores.columns.to_flat_index().str.join("_")
        if _evals is not None:
            for _name, _expr in _evals.items():
                agg_scores[_name] = agg_scores.eval(_expr)
        agg_scores.rename(columns=_scores, inplace=True)
        agg_scores = agg_scores[_scores.values()]
        agg_scores = agg_scores.reset_index().copy()

        def _label_score(score):
            for _label, _thresh in _labels.items():
                if isinstance(_thresh, str):
                    _thresh = eval(_thresh)
                if score >= _thresh[0] and score <= _thresh[1]:
                    return _label
            return np.nan

        for f in _scores.values():
            if f.startswith(feature):
                agg_scores[f] = np.where(
                    agg_scores[_num_examples] > min_examples, agg_scores[f], np.nan
                )
                _label_key = f + "_label"
                if _labels:
                    agg_scores[_label_key] = agg_scores[f].apply(_label_score)

        return agg_scores



import os
import logging
import requests
import pandas as pd
import json
from ekorpkit.io.fetch.base import BaseFetcher
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class LabelStudio(BaseFetcher):
    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        super().__init__(**args)
        self._api = self.args.api
        if not self._api.token:
            raise ValueError(
                "api.token is required, set envrionment variable LABEL_STUDIO_USER_TOKEN or config 'api.token' directly"
            )
        if not self._api.server:
            raise ValueError(
                "api.server is required, set envrionment variable LABEL_STUDIO_SERVER or config 'api.server' directly"
            )
        self._jobs = self.args.jobs
        self._request_schema = self.args.request_schema
        self._project_id = self.args.project_id
        self._json_file = self.args.json_file

    @property
    def project_id(self):
        if not self._project_id:
            raise ValueError("project_id is required")
        return self._project_id

    @property
    def headers(self):
        return eKonf.to_dict(self._api.header)

    def get_url(self, job_name, project_id=None):
        if project_id:
            self._project_id = project_id

        url = self._api.server + self._jobs[job_name]
        if "{id}" in url:
            url = url.format(id=self.project_id)
        return url

    def get_request_schema(self, request_name):
        return eKonf.to_dict(self._request_schema[request_name])

    def list_projects(self, verbose=False):
        url = self.get_url("project")
        response = requests.get(url, headers=self.headers)
        if response.status_code == 200:
            response = response.json()
            if verbose:
                for p in response["results"]:
                    print(f"{p['id']}: {p['title']}")
        return response

    def get_project(self, project_id):
        url = self.get_url("get_project", project_id=project_id)
        response = requests.get(url, headers=self.headers)
        if response.status_code == 200:
            print(f"{project_id} is fetched")
        return response.json()

    def delete_project(self, project_id):
        url = self.get_url("get_project", project_id=project_id)
        response = requests.delete(url, headers=self.headers)
        if response.status_code == 204:
            print(f"{project_id} is deleted")
        else:
            print(f"{project_id} is not deleted")

    def create_project(self, title, description=None, label_config=None, **kargs):
        project_list = self.list_projects()
        for p in project_list["results"]:
            if p["title"] == title:
                print(f"{title} already exists with id {p['id']}")
                return p

        url = self.get_url("project")
        req_body = self.get_request_schema("project")
        req_body["title"] = title
        if description:
            req_body["description"] = description
        if label_config:
            req_body["label_config"] = label_config
        req_body.update(kargs)

        response = requests.post(url, headers=self.headers, json=req_body)
        response = response.json()
        if "id" in response:
            self._project_id = response["id"]
            print(f"project {title} is created with id {self._project_id}")
        else:
            print(f"project {title} is not created")
        return response

    def import_data(self, data, project_id=None):
        url = self.get_url("import", project_id=project_id)
        response = requests.post(url, headers=self.headers, json=data)
        if response.status_code == 201:
            print(f"data is imported to {project_id}")
        return response.json()

    def import_file(self, file_path, project_id=None):
        url = self.get_url("import", project_id=project_id)
        response = requests.post(
            url, headers=self.headers, files={"file": open(file_path, "rb")}
        )
        if response.status_code == 201:
            print(f"{file_path} is imported to {project_id}")
        return response.json()

    def download_export(self, project_id=None, json_file=None):
        req_body = self.get_request_schema("export")
        url = self.get_url("export", project_id=project_id)

        json_file = json_file or self._json_file
        json_path = os.path.join(self.output_dir, json_file)
        log.info(f"fetching {url} with {req_body} to {json_path}")
        response = requests.get(url, headers=self.headers, params=req_body)
        with open(json_path, "w") as f:
            f.write(response.text)
        log.info(f"{json_path} is downloaded")
        return json_path

    def annotations_to_dataframe(self, json_path, output_file=None):

        df = pd.read_json(json_path)

        def to_list(annotations):
            return [
                [
                    ann["id"],
                    ann["completed_by"],
                    ann["result"][0]["origin"],
                    ann["result"][0]["value"]["choices"][0],
                ]
                for ann in annotations
                if ann["result"]
            ]

        df["annotations"] = df.annotations.apply(to_list)
        df["text"] = df.data.apply(lambda x: x["text"])
        cols = ["id", "annotations", "text"]
        df = df[cols]
        df = df.explode("annotations")
        cols = ["annot_id", "annotator", "origin", "labels"]
        df[cols] = df.annotations.apply(pd.Series)
        df = df.dropna(subset=cols)
        cols = ["annot_id", "annotator"]
        df[cols] = df[cols].astype(int)
        df = df.drop(columns="annotations")
        self._data = df
        if self.verbose:
            print(df.head())
        output_file = output_file or self._output_file
        eKonf.save_data(df, output_file)
        print(f"{output_file} is saved")
        return df

    def dataframe_to_predictions(
        self,
        data,
        json_file="predictions.json",
        type="choices",
        choices_name="sentiment",
        choices_toname="text",
        data_choices="pred_labels",
        data_score="pred_probs",
        model_version="1.0",
    ):
        data = data.to_dict(orient="records")
        annotations = []
        for d in data:
            annotations.append(
                {
                    "data": d,
                    "predictions": [
                        {
                            "model_version": model_version,
                            "result": [
                                {
                                    "from_name": choices_name,
                                    "to_name": choices_toname,
                                    "type": type,
                                    "readonly": False,
                                    "hidden": False,
                                    "value": {
                                        "choices": [d[data_choices]],
                                    },
                                }
                            ],
                            "score": d[data_score],
                        }
                    ],
                }
            )
        json_path = os.path.join(self.output_dir, json_file)
        with open(json_path, "w") as f:
            json.dump(annotations, f, ensure_ascii=False)
        print(f"{json_path} is saved")
        return json_path



import logging

log = logging.getLogger(__name__)


def html_to_json(contents):
    import html_to_json

    doc = {}
    output_json = html_to_json.convert(contents)
    for k, v in output_json.items():
        if isinstance(v, list) and len(v) == 1:
            doc[k] = v[0].get("_value")
    return [doc]


def parse_plaintext(
    contents,
    split=False,
    meta_line=None,
    meta_key="meta",
    data_key="text",
    lineno_key="lineno",
    progress_per=1000,
    verbose=False,
    **kwargs,
):
    if split:
        docs = []
        for i, line in enumerate(contents.splitlines()):
            line = line.strip()
            doc = {lineno_key: i, data_key: line}
            if i < 2 and verbose:
                log.info(f"processing line {i}")
                log.info(doc)
            elif progress_per and i > 0 and i % progress_per == 0:
                log.info(f"processing line {i}")
            docs.append(doc)
        return docs
    elif meta_line is not None:
        text = []
        meta = []
        for i, line in enumerate(contents.splitlines()):
            line = line.strip()
            if i < meta_line:
                meta.append(line)
            else:
                text.append(line)
        doc = {meta_key: "\n".join(meta), data_key: "\n".join(text)}
        return [doc]
    else:
        doc = {data_key: "\n".join(contents.splitlines())}
        return [doc]


def parse_reuters_contents(contents):
    text = ""
    title = ""
    author = ""
    date = ""
    url = ""

    for i, line in enumerate(contents.split("\n")):
        if i == 0 and line.startswith("-- "):
            title = line.replace("-- ", "").strip()
        elif i == 1 and line.startswith("-- "):
            author = line.replace("-- ", "").replace("By ", "").strip()
        elif i == 2 and line.startswith("-- "):
            date = line.replace("-- ", "").strip()
        elif i == 3 and line.startswith("-- "):
            url = line.replace("-- ", "").strip()
        else:
            text += line + "\n"
    doc = {
        "title": title,
        "author": author,
        "date": date,
        "url": url,
        "text": text.strip(),
    }
    return [doc]



#!python
from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
import pandas as pd
from .batcher import Batcher

# from tqdm.auto import tqdm


def decorator_apply(
    func, batcher=None, cache=None, vectorize=None, description="batch_apply"
):
    def wrapper_func(*args, **kwargs):
        return Apply(
            func,
            args=args[1:],
            kwargs=kwargs,
            batcher=batcher,
            cache=cache,
            vectorize=vectorize,
            description=description,
        ).transform(args[0])

    return wrapper_func


def batch_transform(args):
    f = args[1]
    f_args = args[2]
    f_kwargs = args[3]
    if args[5] is not None:
        from numba import vectorize

        return vectorize(args[5], fastmath=True)(f)(*zip(*args[0]))
    if args[4] is not None:
        from functools import lru_cache

        f = lru_cache(maxsize=args[4])(f)
    # Applying per DataFrame row is very slow, use ApplyBatch instead
    if isinstance(args[0], pd.DataFrame):
        return args[0].apply(lambda x: f(x, *f_args, **f_kwargs), axis=1)
    elif isinstance(args[0], pd.Series):
        return args[0].apply(lambda x: f(x, *f_args, **f_kwargs))
    return [f(row, *f_args, **f_kwargs) for row in args[0]]


class Apply(object):
    # Applies a function to each row of a minibatch
    def __init__(
        self,
        function,
        batcher=None,
        args=[],
        kwargs={},
        cache=None,
        vectorize=None,
        description="batch_apply",
    ):
        if batcher is None:
            self.batcher = Batcher()
        else:
            self.batcher = batcher
        self.function = function
        self.args = [args]
        self.kwargs = [kwargs]
        self.cache = [cache]
        self.vectorize = [vectorize]
        self.description = description

    def fit(self, data, input_split=False, batcher=None):
        return self

    def fit_transform(
        self,
        data,
        input_split=False,
        merge_output=True,
        minibatch_size=None,
        batcher=None,
    ):
        return self.transform(data, input_split, merge_output, minibatch_size, batcher)

    def transform(
        self,
        data,
        input_split=False,
        merge_output=True,
        minibatch_size=None,
        batcher=None,
    ):
        if batcher is None:
            batcher = self.batcher
        return batcher.process_batches(
            batch_transform,
            data,
            [self.function] + self.args + self.kwargs + self.cache + self.vectorize,
            input_split=input_split,
            merge_output=merge_output,
            minibatch_size=minibatch_size,
            description=self.description,
        )


# import wordbatch.batcher as batcher
# b= batcher.Batcher(minibatch_size=2)#, method="serial")
# import numpy as np
# a= Apply(np.power, b, [2],{})
# print(a.transform([1, 2, 3, 4]))



import os
import logging
from abc import ABCMeta
from sklearn import preprocessing
from ekorpkit.pipelines.pipe import apply_pipeline
from ekorpkit import eKonf


DESCRIPTION = "ekorpkit datasets"
LICENSE = "Copyright of the dataset is owned by the authors."


log = logging.getLogger(__name__)


class BaseSet:
    __metaclass__ = ABCMeta

    SPLITS = eKonf.SPLITS

    def __init__(self, **args):
        args = eKonf.to_config(args)
        self.args = args
        self.name = self.args.name
        self.verbose = self.args.get("verbose", False)
        self.auto = self.args.auto
        self.force = self.args.force
        self.data_dir = self.args["data_dir"]
        self._collapse_ids = self.args.get("collapse_ids", False)

        self._info_cfg = self.args.get("info", None)
        self._pipeline_cfg = self.args.get("pipeline") or {}
        self._pipeline_ = self._pipeline_cfg.get(eKonf.Keys.PIPELINE, [])
        if self._pipeline_ is None:
            self._pipeline_ = []

        self.filetype = None
        self.data_files = None
        self._info = None
        self._summary_info = None
        self._column = None
        self._splits = {}
        self._data = None
        self._loaded = False
        self._classes = None
        self._le = None

    def load_info(self):
        """Load the info file."""
        self.info_file = os.path.join(self.data_dir, f"info-{self.name}.yaml")
        self._info = (
            eKonf.load(self.info_file)
            if eKonf.exists(self.info_file) and not self.force.build
            else {}
        )
        if self._info:
            log.info(f"Loaded info file: {self.info_file}")
            self.args = eKonf.merge(self.args, self._info)
            self._info = eKonf.to_dict(self._info)
        self.filetype = self.args.get("filetype") or "parquet"
        self.filetype = "." + self.filetype.replace(".", "")
        self.description = self.args.get("description", DESCRIPTION)
        self.license = self.args.get("license", LICENSE)
        if self.verbose:
            log.info(
                f"Intantiating a {self.__class__.__name__} [{self.name}] with a config:"
            )
            eKonf.pprint(self.args)
        self.data_files = self.args.get("data_files", None)
        if self.data_files is None:
            self.data_files = {
                self.SPLITS.TRAIN.value: f"{self.name}-train{self.filetype}",
                self.SPLITS.DEV.value: f"{self.name}-dev{self.filetype}",
                self.SPLITS.TEST.value: f"{self.name}-test{self.filetype}",
            }

    def load_column_info(self):
        self._column_info = self.args.get("column_info")
        if self._column_info is None:
            raise ValueError("Column info can't be None")
        self._column = eKonf.instantiate(self._column_info)

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname} : {self.name}"
        return s

    def __getitem__(self, split="train"):
        if split in self.splits:
            return self.splits[split]
        else:
            return None

    def __len__(self):
        return self.num_rows

    @property
    def num_rows(self) -> int:
        """Number of rows in the corpus (same as :meth:`Corpus.__len__`)."""
        if self.data.index is not None:
            return len(self.data.index)
        return len(self.data)

    @property
    def INFO(self):
        return self._info

    @property
    def COLUMN(self):
        return self._column

    @property
    def ID(self):
        return self.COLUMN.ID

    @property
    def IDs(self):
        return self.COLUMN.IDs

    @property
    def DATA(self):
        return self.COLUMN.DATA

    @property
    def DATATYPEs(self):
        return self.COLUMN.DATATYPEs

    @property
    def data(self):
        dfs = []
        for split, _data in self.splits.items():
            if _data is not None:
                dfs.append(_data)
        df = eKonf.concat_data(dfs)
        return df

    @property
    def splits(self):
        return self._splits

    @property
    def summary_info(self):
        return self._summary_info

    @property
    def classes(self):
        if self._classes is None:
            log.info("LabelEncoder is not fitted")
            return None
        return self._classes.tolist()

    def build(self):
        data = None
        if self._pipeline_ and len(self._pipeline_) > 0:
            data = apply_pipeline(data, self._pipeline_, self._pipeline_cfg)
        if data is not None:
            log.info(f"Dataset {self.name} built with {len(data)} rows")
        else:
            log.info(f"Dataset {self.name} is empty")

    def persist(self):
        if not self._loaded:
            log.info(f"Dataset {self.name} is not loaded")
            return
        if self.summary_info is None:
            self.summarize()
        for split, data in self._splits.items():
            if data is None:
                continue
            data_file = self.data_files[split]
            eKonf.save_data(
                data,
                data_file,
                base_dir=self.data_dir,
                verbose=self.verbose,
            )
        if self.summary_info is not None:
            self.summary_info.save(info={"column_info": self.COLUMN.INFO})

    def save_as(self, name):
        if not self._loaded:
            log.info(f"Dataset {self.name} is not loaded")
            return
        self.data_dir = self.data_dir.replace(self.name, name)
        self.name = name
        self.args.data_files = None
        self._info_cfg.data_dir = self.data_dir
        self._info_cfg.name = self.name
        self._summary_info = None
        self.load_info()
        self.persist()

    def load(self):
        if self._loaded:
            return
        for split, data_file in self.data_files.items():
            if eKonf.exists(self.data_dir, data_file):
                data = eKonf.load_data(
                    data_file,
                    self.data_dir,
                    verbose=self.verbose,
                    concatenate=True,
                )
                data = self.COLUMN.init_info(data)
                data = self.COLUMN.append_split(data, split)
                if self._collapse_ids:
                    data = self.COLUMN.combine_ids(data)
                self._splits[split] = data
                if self.verbose:
                    log.info(f"Data loaded {len(data)} rows")
                    print(data.head(3))
                    print(data.tail(3))
            else:
                log.warning(f"File {data_file} not found.")
                # log.info(f"Dataset {self.name} split {split} is empty")
        self._loaded = True

    def summarize(self):
        if not self._loaded:
            log.info(f"Dataset {self.name} is not loaded")
            return
        summary_info = None
        if self._info_cfg:
            summary_info = eKonf.instantiate(self._info_cfg)
        if summary_info:
            summary_info.load(self.INFO)
        for split, data in self.splits.items():
            if data is None:
                continue
            data_file = self.data_files[split]
            if summary_info:
                stats = {"data_file": data_file}
                summary_info.init_stats(split_name=split, stats=stats)
                summary_info.calculate_stats(data, split)
        self._summary_info = summary_info

    def fit_labelencoder(self, data):
        self._le = preprocessing.LabelEncoder()
        self._le.fit(data)
        self._classes = self._le.classes_
        log.info(f"LabelEncoder classes: {self._classes}")

    def transform_labels(self, data):
        if not self._loaded:
            log.info(f"Dataset {self.name} is not loaded")
            return data
        if data is None:
            log.info("Data is None")
            return data
        if self._le is None:
            log.info("Label encoder is not fitted")
            self.fit_labelencoder(data)
        _data = self._le.transform(data)
        return _data

    def inverse_transform_labels(self, data):
        if not self._loaded:
            log.info(f"Dataset {self.name} is not loaded")
            return data
        if data is None:
            log.info("Data is None")
            return data
        if self._le is None:
            log.info("Label encoder is not fitted")
            self.fit_labelencoder(data)
        _data = self._le.inverse_transform(data)
        return _data



import pandas as pd
import logging
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from .corpus import Corpus
from .base import BaseSet

log = logging.getLogger(__name__)


class Corpora(BaseSet):
    def __init__(self, **args):
        super().__init__(**args)
        self.corpora = args.get("corpora")
        self.corpora = eKonf.to_dict(self.corpora)
        if self.corpora is None:
            self.corpora = self.name
        if isinstance(self.corpora, str):
            self.corpora = {self.corpora: None}
        elif eKonf.is_list(self.corpora):
            self.corpora = {name: None for name in self.corpora}
        # if eKonf.is_list(self.name):
        #     self.name = "-".join(self.name)
        if self.name is None and eKonf.is_dict(self.corpora):
            self.name = "-".join(self.corpora.keys())

        self.metadata_dir = self.args.get("metadata_dir", None)
        if self.metadata_dir is None:
            self.metadata_dir = self.data_dir
        self.meta_files = self.args.get("meta_files", None)
        use_name_as_subdir = args.get("use_name_as_subdir", True)

        self.load_column_info()

        self._metadata = None

        with elapsed_timer(format_time=True) as elapsed:
            for name in self.corpora:
                log.info(f"processing {name}")
                args["name"] = name
                args["data_dir"] = self.data_dir
                args["metadata_dir"] = self.metadata_dir
                args["auto"] = self.auto
                args["use_name_as_subdir"] = use_name_as_subdir
                if self.data_files is not None:
                    if name in self.data_files:
                        args["data_files"] = self.data_files[name]
                    elif "train" in self.data_files:
                        args["data_files"] = self.data_files
                if self.meta_files is not None:
                    if name in self.meta_files:
                        args["meta_files"] = self.meta_files[name]
                    elif "train" in self.meta_files:
                        args["meta_files"] = self.meta_files
                corpus = Corpus(**args)
                self.corpora[name] = corpus
            log.info(f">>> Elapsed time: {elapsed()} <<< ")

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname}\n----------\n"
        for name in self.corpora.keys():
            s += f"{str(name)}\n"
        return s

    def __getitem__(self, name):
        return self.corpora[name]

    @property
    def TEXT(self):
        return self.COLUMN.TEXT

    @property
    def data(self):
        return self._data

    @property
    def metadata(self):
        return self._metadata

    def load(self):
        for _name in self.corpora:
            self.corpora[_name].load()
        self._loaded = True

    def concat_corpora(self, append_corpus_name=True):
        self.concatenate(append_name=append_corpus_name)

    def concatenate(self, append_name=True):
        if not self._loaded:
            self.load()

        dfs, df_metas = [], []

        for name in self.corpora:
            df = self.corpora[name].data
            if df is None:
                self.load()
            if append_name:
                df = self.COLUMN.append_corpus(df, name)
            dfs.append(df)
            df_meta = self.corpora[name]._metadata
            if df_meta is not None:
                if append_name:
                    df_meta = self.COLUMN.append_corpus(df_meta, name)
                df_metas.append(df_meta)
        self._data = pd.concat(dfs, ignore_index=True)
        if len(df_metas) > 0:
            self._metadata = pd.concat(df_metas, ignore_index=True)
        if self.verbose:
            log.info(f"concatenated {len(dfs)} corpora")
            print(self._data.head())

    def __iter__(self):
        for corpus in self.corpora.values():
            yield corpus

    def __getitem__(self, name):
        if name not in self.corpora:
            raise KeyError(f"{name} not in corpora")
        return self.corpora[name]

    def __len__(self):
        return len(self.corpora)



import logging
import matplotlib.pyplot as plt
import seaborn as sns
from cmath import isinf
from pathlib import Path
from .base import (
    set_style,
    set_figure,
    set_super,
    add_decorations,
    save_figure,
    prepare_plot_args,
    prepare_data,
)
from .classification import confusion_matrix
from .yellowbrick import yellowbrick_features
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def plot(data, verbose=False, **kwargs):
    (
        _axes,
        _fig_super,
        _,
        _gridspec,
        _plots,
        _savefig,
        _style,
        _subplots,
        _,
        data,
        figsize,
        tight_layout,
    ) = prepare_plot_args(data, verbose, **kwargs)

    set_style(**_style)

    if verbose and isinstance(_plots, list) and len(_plots) > 1:
        log.info("Plotting multiple plots'")

    sencondary_axes = {}
    fig = plt.figure(figsize=figsize, tight_layout=tight_layout)

    gs = fig.add_gridspec(**_gridspec)
    axes = gs.subplots(**_subplots)
    if _gridspec["nrows"] > 1 or _gridspec["ncols"] > 1:
        ax = None
        axes = axes.flatten()
    else:
        axes = [axes]
        ax = axes[0]

    if isinstance(_plots, dict):
        _plots = [_plots]
    if not isinstance(_plots, list) or data is None:
        log.info("No plots to plot")
        _plots = []
    for i, _plot_cfg_ in enumerate(_plots):
        _func_ = eval(_plot_cfg_.pop(eKonf.Keys.FUNC))
        _x = _plot_cfg_.pop("x", None)
        _y = _plot_cfg_.pop("y", None)
        _secondary_y = _plot_cfg_.pop("secondary_y", False)
        _secondary_to = _plot_cfg_.pop("secondary_to", 0)
        _axno = _plot_cfg_.pop("axno", 0)
        _datano = _plot_cfg_.pop("datano", 0)
        _query = _plot_cfg_.pop("query", None)
        _set_index = _plot_cfg_.pop("set_index", None)

        if _secondary_y:
            if _secondary_to in sencondary_axes:
                ax = sencondary_axes[_secondary_to]
            else:
                ax = axes[_secondary_to].twinx()
                log.info(f"Creating secondary axis to axis[{_secondary_to}]")
                sencondary_axes[_secondary_to] = ax
        else:
            ax = axes[_axno]
        if isinstance(data, list):
            _data = data[_datano]
            log.debug(f"Plotting data[{_datano}]")
        else:
            _data = data
        _data = prepare_data(_data, _query, _set_index)
        _func_(ax, _x, _y, _data, **_plot_cfg_)

    if _axes is None:
        _axes = []
    elif not isinstance(_axes, (dict, list)):
        _axes = []
    elif isinstance(_axes, dict):
        _axes = [_axes]
    for _ax_cfg_ in _axes:
        _secondary_y = _ax_cfg_.get("secondary_y", False)
        if _secondary_y:
            _secondary_to = _ax_cfg_.get("secondary_to", 0)
            if _secondary_to in sencondary_axes:
                ax = sencondary_axes[_secondary_to]
            else:
                ax = None
        else:
            _axno = _ax_cfg_.get("axno", 0)
            ax = axes[_axno]
        add_decorations(ax, **_ax_cfg_)
        set_figure(ax, **_ax_cfg_)

    set_super(fig, **_fig_super)

    save_figure(fig, **_savefig)


def grid(data, verbose=False, **kwargs):
    (
        _,
        _,
        _grid,
        _,
        _,
        _savefig,
        _style,
        _,
        _theme,
        data,
        figsize,
        _,
    ) = prepare_plot_args(data, verbose, **kwargs)

    set_style(**_style)

    if isinstance(_theme["rc"], dict):
        _theme["rc"]["figure.figsize"] = figsize
    else:
        _theme["rc"] = {"figure.figsize": figsize}
    sns.set_theme(**_theme)

    _query = _grid.pop("query", None)
    _set_index = _grid.pop("set_index", None)
    _func_ = eval(_grid.get(eKonf.Keys.FUNC))
    _data = prepare_data(data, _query, _set_index)
    g = _func_(_data, **_grid)

    save_figure(plt, **_savefig)


def facetgrid(data=None, **kwargs):
    rcParams = {} or kwargs.get(eKonf.Keys.rcPARAMS)
    _name_ = kwargs.get(eKonf.Keys.METHOD_NAME)
    _map = kwargs.get(_name_)
    _map_fn = _map.get(eKonf.Keys.FUNC)
    _map_parms = _map.get(eKonf.Keys.rcPARAMS)

    g = sns.FacetGrid(data, **rcParams)
    map_fn = getattr(sns, _map_fn)
    getattr(g, _name_)(map_fn, **_map_parms)
    if kwargs.get("add_legend", False):
        g.add_legend()
    return g


def snsplot(ax=None, x=None, y=None, data=None, **kwargs):
    rcParams = kwargs.pop(eKonf.Keys.rcPARAMS, {}) or {}
    _name_ = kwargs.pop(eKonf.Keys.METHOD_NAME)
    if ax is None:
        ax = plt.gca()
    if isinstance(y, list) and len(y) == 1:
        y = y[0]
    if isinstance(y, list):
        if x is not None and x in data.columns:
            data.set_index(x, inplace=True)
        data = data[y]
        x = None
        y = None
    elif isinstance(y, str):
        if x is None:
            if data.index.name is not None:
                x = data.index.name
            else:
                x = data.index
    if x is not None:
        rcParams["x"] = x
    if y is not None:
        rcParams["y"] = y
    _fn = getattr(sns, _name_)
    if isinstance(kwargs, dict):
        rcParams.update(kwargs)
    log.info(f"Plotting {_name_} with {rcParams}")
    _fn(data=data, ax=ax, **rcParams)


def heatmap(ax=None, x=None, y=None, data=None, **kwargs):
    rcParams = {} or kwargs.get(eKonf.Keys.rcPARAMS)
    if ax is None:
        ax = plt.gca()
    sns.heatmap(data=data, ax=ax, **rcParams)


def stackplot(ax=None, x=None, y=None, data=None, **kwargs):
    rcParams = {} or kwargs.get(eKonf.Keys.rcPARAMS)
    if ax is None:
        ax = plt.gca()
    if x is None or x not in data.columns:
        x = data.index
    elif isinstance(x, str):
        x = data[x]
    plt.stackplot(x, data[y].T, **rcParams)



import re
import collections


class WordPieceTokenizer:
    def __init__(
        self,
        suffix_indicator="##",
        special_tokens=["[PAD]", "[UNK]", "[CLS]", "[SEP]", "[MASK]"],
        unknown_token="[UNK]",
        lowercase=True,
    ):
        self.vocab_tokenized = {}

        self.word_freqs = {}
        self.characters = []
        self.vocab = []
        self.initial_splits = {}

        self.suffix_indicator = suffix_indicator
        self.special_tokens = special_tokens
        self.unknown_token = unknown_token
        self.lowercase = lowercase

    def format_word(self, word):
        return " ".join(list(word))

    def initialize_vocab(self, texts):
        self.word_freqs = self.get_word_frequency(texts)
        self.characters = self.initialize_characters(self.word_freqs)
        self.initial_splits = self.split_words(self.word_freqs)
        self.vocab = self.special_tokens + self.characters.copy()

    def get_word_frequency(self, texts):
        word_freqs = {}
        for text in texts:
            words = self.pre_tokenize(text)
            for word in words:
                word_freqs[word] = word_freqs.get(word, 0) + 1
        return word_freqs

    def initialize_characters(self, word_freqs):
        characters = []
        for word in word_freqs.keys():
            if word[0] not in characters:
                characters.append(word[0])
            for letter in word[1:]:
                if f"{self.suffix_indicator}{letter}" not in characters:
                    characters.append(f"{self.suffix_indicator}{letter}")
        return characters

    def split_words(self, word_freqs):
        splits = {
            word: [
                c if i == 0 else f"{self.suffix_indicator}{c}"
                for i, c in enumerate(word)
            ]
            for word in word_freqs.keys()
        }
        return splits

    def compute_pair_scores(self, splits):
        letter_freqs = collections.defaultdict(int)
        pair_freqs = collections.defaultdict(int)
        for word, freq in self.word_freqs.items():
            split = splits[word]
            if len(split) == 1:
                letter_freqs[split[0]] += freq
                continue
            for i in range(len(split) - 1):
                pair = (split[i], split[i + 1])
                letter_freqs[split[i]] += freq
                pair_freqs[pair] += freq
            letter_freqs[split[-1]] += freq

        scores = {
            pair: freq / (letter_freqs[pair[0]] * letter_freqs[pair[1]])
            for pair, freq in pair_freqs.items()
        }
        return scores

    def merge_pair(self, a, b, splits):
        for word in self.word_freqs:
            split = splits[word]
            if len(split) == 1:
                continue
            i = 0
            while i < len(split) - 1:
                if split[i] == a and split[i + 1] == b:
                    merge = a + b[2:] if b.startswith(self.suffix_indicator) else a + b
                    split = split[:i] + [merge] + split[i + 2 :]
                else:
                    i += 1
            splits[word] = split
        return splits

    def fit(self, texts, vocab_size, verbose=True, print_every=100):
        self.initialize_vocab(texts)
        i = 0
        splits = self.initial_splits.copy()
        while len(self.vocab) < vocab_size:
            scores = self.compute_pair_scores(splits)
            best_pair, max_score = "", None
            for pair, score in scores.items():
                if max_score is None or max_score < score:
                    best_pair = pair
                    max_score = score
            splits = self.merge_pair(*best_pair, splits)
            new_token = (
                best_pair[0] + best_pair[1][2:]
                if best_pair[1].startswith(self.suffix_indicator)
                else best_pair[0] + best_pair[1]
            )
            self.vocab.append(new_token)
            if verbose and print_every > 0 and i % print_every == 0:
                print("Iteration: {}, vocab size: {}".format(i, len(self.vocab)))
                print("Best pair: {}".format(best_pair))
                print("New token: {}".format(new_token))
            i += 1
        if verbose:
            print("Final vocab size: {}".format(len(self.vocab)))

    def encode_word(self, word):
        if word in self.vocab_tokenized:
            return self.vocab_tokenized[word]
        else:
            tokens = []
            while len(word) > 0:
                i = len(word)
                while i > 0 and word[:i] not in self.vocab:
                    i -= 1
                if i == 0:
                    return [self.unknown_token]
                tokens.append(word[:i])
                word = word[i:]
                if len(word) > 0:
                    word = f"{self.suffix_indicator}{word}"
            self.vocab_tokenized[word] = tokens
            return tokens

    def print_enocoded_word(self, word):
        print("Encoding word: {}...".format(word))
        if word in self.vocab_tokenized:
            print("Encoding of the known word:")
            print(self.vocab_tokenized[word])
            print("Encoding treating the known word as unknown:")
            print(self.encode_word(word))
        else:
            print("Encoding of the unknown word:")
            print(self.encode_word(word))

    def pre_tokenize(self, text):
        if self.lowercase:
            text = text.lower()
        text = re.sub(r"\s+", " ", text)
        return text.split()

    def tokenize(self, text):
        words = self.pre_tokenize(text)
        encoded_words = [self.encode_word(word) for word in words]
        return sum(encoded_words, [])



import logging
import torch, torchvision
import py3d_tools as p3d
import midas_utils
from PIL import Image
import numpy as np
import math

try:
    from infer import InferenceHelper
except:
    print(
        "disco_xform_utils.py failed to import InferenceHelper. Please ensure that AdaBins directory is in the path (i.e. via sys.path.append('./AdaBins') or other means)."
    )

MAX_ADABINS_AREA = 500000
MIN_ADABINS_AREA = 448 * 448


log = logging.getLogger(__name__)


@torch.no_grad()
def transform_image_3d(
    img_filepath,
    midas_model,
    midas_transform,
    device,
    rot_mat=torch.eye(3).unsqueeze(0),
    translate=(0.0, 0.0, -0.04),
    near=2000,
    far=20000,
    fov_deg=60,
    padding_mode="border",
    sampling_mode="bicubic",
    midas_weight=0.3,
    spherical=False,
):
    img_pil = Image.open(open(img_filepath, "rb")).convert("RGB")
    w, h = img_pil.size
    image_tensor = torchvision.transforms.functional.to_tensor(img_pil).to(device)

    use_adabins = midas_weight < 1.0

    if use_adabins:
        # AdaBins
        """
        predictions using nyu dataset
        """
        log.info("Running AdaBins depth estimation implementation...")
        infer_helper = InferenceHelper(dataset="nyu", device=device)

        image_pil_area = w * h
        if image_pil_area > MAX_ADABINS_AREA:
            scale = math.sqrt(MAX_ADABINS_AREA) / math.sqrt(image_pil_area)
            depth_input = img_pil.resize(
                (int(w * scale), int(h * scale)), Image.LANCZOS
            )  # LANCZOS is supposed to be good for downsampling.
        elif image_pil_area < MIN_ADABINS_AREA:
            scale = math.sqrt(MIN_ADABINS_AREA) / math.sqrt(image_pil_area)
            depth_input = img_pil.resize(
                (int(w * scale), int(h * scale)), Image.BICUBIC
            )
        else:
            depth_input = img_pil
        try:
            _, adabins_depth = infer_helper.predict_pil(depth_input)
            if image_pil_area != MAX_ADABINS_AREA:
                adabins_depth = (
                    torchvision.transforms.functional.resize(
                        torch.from_numpy(adabins_depth),
                        image_tensor.shape[-2:],
                        interpolation=torchvision.transforms.functional.InterpolationMode.BICUBIC,
                    )
                    .squeeze()
                    .to(device)
                )
            else:
                adabins_depth = torch.from_numpy(adabins_depth).squeeze().to(device)
            adabins_depth_np = adabins_depth.cpu().numpy()
        except:
            pass

    torch.cuda.empty_cache()

    # MiDaS
    img_midas = midas_utils.read_image(img_filepath)
    img_midas_input = midas_transform({"image": img_midas})["image"]
    midas_optimize = True

    # MiDaS depth estimation implementation
    log.info("Running MiDaS depth estimation implementation...")
    sample = torch.from_numpy(img_midas_input).float().to(device).unsqueeze(0)
    if midas_optimize == True and device == torch.device("cuda"):
        sample = sample.to(memory_format=torch.channels_last)
        sample = sample.half()
    prediction_torch = midas_model.forward(sample)
    prediction_torch = torch.nn.functional.interpolate(
        prediction_torch.unsqueeze(1),
        size=img_midas.shape[:2],
        mode="bicubic",
        align_corners=False,
    ).squeeze()
    prediction_np = prediction_torch.clone().cpu().numpy()

    log.info("Finished depth estimation.")
    torch.cuda.empty_cache()

    # MiDaS makes the near values greater, and the far values lesser. Let's reverse that and try to align with AdaBins a bit better.
    prediction_np = np.subtract(50.0, prediction_np)
    prediction_np = prediction_np / 19.0

    if use_adabins:
        adabins_weight = 1.0 - midas_weight
        depth_map = prediction_np * midas_weight + adabins_depth_np * adabins_weight
    else:
        depth_map = prediction_np

    depth_map = np.expand_dims(depth_map, axis=0)
    depth_tensor = torch.from_numpy(depth_map).squeeze().to(device)

    pixel_aspect = 1.0  # really.. the aspect of an individual pixel! (so usually 1.0)
    persp_cam_old = p3d.FoVPerspectiveCameras(
        near, far, pixel_aspect, fov=fov_deg, degrees=True, device=device
    )
    persp_cam_new = p3d.FoVPerspectiveCameras(
        near,
        far,
        pixel_aspect,
        fov=fov_deg,
        degrees=True,
        R=rot_mat,
        T=torch.tensor([translate]),
        device=device,
    )

    # range of [-1,1] is important to torch grid_sample's padding handling
    y, x = torch.meshgrid(
        torch.linspace(-1.0, 1.0, h, dtype=torch.float32, device=device),
        torch.linspace(-1.0, 1.0, w, dtype=torch.float32, device=device),
    )
    z = torch.as_tensor(depth_tensor, dtype=torch.float32, device=device)
    xyz_old_world = torch.stack((x.flatten(), y.flatten(), z.flatten()), dim=1)

    # Transform the points using pytorch3d. With current functionality, this is overkill and prevents it from working on Windows.
    # If you want it to run on Windows (without pytorch3d), then the transforms (and/or perspective if that's separate) can be done pretty easily without it.
    xyz_old_cam_xy = persp_cam_old.get_full_projection_transform().transform_points(
        xyz_old_world
    )[:, 0:2]
    xyz_new_cam_xy = persp_cam_new.get_full_projection_transform().transform_points(
        xyz_old_world
    )[:, 0:2]

    offset_xy = xyz_new_cam_xy - xyz_old_cam_xy
    # affine_grid theta param expects a batch of 2D mats. Each is 2x3 to do rotation+translation.
    identity_2d_batch = torch.tensor(
        [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]], device=device
    ).unsqueeze(0)
    # coords_2d will have shape (N,H,W,2).. which is also what grid_sample needs.
    coords_2d = torch.nn.functional.affine_grid(
        identity_2d_batch, [1, 1, h, w], align_corners=False
    )
    offset_coords_2d = coords_2d - torch.reshape(offset_xy, (h, w, 2)).unsqueeze(0)

    if spherical:
        spherical_grid = get_spherical_projection(
            h, w, torch.tensor([0, 0], device=device), -0.4, device=device
        )  # align_corners=False
        stage_image = torch.nn.functional.grid_sample(
            image_tensor.add(1 / 512 - 0.0001).unsqueeze(0),
            offset_coords_2d,
            mode=sampling_mode,
            padding_mode=padding_mode,
            align_corners=True,
        )
        new_image = torch.nn.functional.grid_sample(
            stage_image, spherical_grid, align_corners=True
        )  # , mode=sampling_mode, padding_mode=padding_mode, align_corners=False)
    else:
        new_image = torch.nn.functional.grid_sample(
            image_tensor.add(1 / 512 - 0.0001).unsqueeze(0),
            offset_coords_2d,
            mode=sampling_mode,
            padding_mode=padding_mode,
            align_corners=False,
        )

    img_pil = torchvision.transforms.ToPILImage()(new_image.squeeze().clamp(0, 1.0))

    torch.cuda.empty_cache()

    return img_pil


def get_spherical_projection(H, W, center, magnitude, device):
    xx, yy = torch.linspace(
        -1, 1, W, dtype=torch.float32, device=device
    ), torch.linspace(-1, 1, H, dtype=torch.float32, device=device)
    gridy, gridx = torch.meshgrid(yy, xx)
    grid = torch.stack([gridx, gridy], dim=-1)
    d = center - grid
    d_sum = torch.sqrt((d ** 2).sum(axis=-1))
    grid += d * d_sum.unsqueeze(-1) * magnitude
    return grid.unsqueeze(0)



import codecs
import logging
import pandas as pd
from ekorpkit import eKonf

log = logging.getLogger(__name__)


class BaseInfo:

    Keys = eKonf.Keys

    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        self._initialized = False

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname} :\n{self.INFO}"
        return s

    def init_info(self, data):
        if self._initialized:
            return data
        if isinstance(data, pd.DataFrame):
            log.info(
                f"index: {self.INDEX}, index of data: {data.index.name}, columns: {list(data.columns)}, id: {self.IDs}"
            )
            if data.index.name is None:
                data.index.name = self.INDEX
            elif self.INDEX is None:
                self.INDEX = data.index.name
            elif self.INDEX != data.index.name and self.INDEX in data.columns:
                data = self.set_index(data, self.INDEX)
            elif self.INDEX != data.index.name and self.INDEX not in data.columns:
                log.warning(f"{self.INDEX} not in dataframe")

            if not self.IDs or self.IDs[0] == self.Keys.INDEX.value:
                self.IDs = [self.INDEX]
            self.set_dtypes(data)
            self._initialized = True
        return data

    def set_index(self, data, name):
        if isinstance(data, pd.DataFrame):
            if name in data.columns:
                data.set_index(name, inplace=True)
                self.INDEX = name
            else:
                log.warning(f"{name} not in dataframe")
        return data

    def reset_index(
        self,
        data,
        rename_old_index=None,
        drop=False,
    ):
        if isinstance(data, pd.DataFrame):
            if self.INDEX in data.columns:
                data.drop(self.INDEX, axis=1, inplace=True)
            data = data.reset_index(drop=drop)
            if not drop and rename_old_index is not None and self.INDEX in data.columns:
                data = data.rename(columns={self.INDEX: rename_old_index})
            self.INDEX = self.Keys.INDEX.value
            self.set_dtypes(data)
        return data

    def reset_id(self, data):
        if isinstance(data, pd.DataFrame):
            data.rename(columns={self.ID: self._ID}, inplace=True)
            data = self.reset_index(data, rename_old_index=self.ID)
        return data

    def combine_ids(self, data):
        if self.IDs is None:
            return data

        if isinstance(data, pd.DataFrame):
            if len(self.IDS) > 1:
                data[self.ID] = data[self.IDs].apply(
                    lambda row: self.ID_SEPARATOR.join(row.values.astype(str)),
                    axis=1,
                )

        return data

    def common_columns(self, dataframes):
        """
        Find common columns between dataframes
        """
        if not isinstance(dataframes, list):
            dataframes = [dataframes]
        common_columns = list(set.intersection(*(set(df.columns) for df in dataframes)))
        df = dataframes[0][common_columns].copy()
        dtypes = df.dtypes.apply(lambda x: x.name).to_dict()
        self.DATATYPEs = dtypes
        return common_columns

    def to_datetime(self, data):
        if self.DATETIME_INFO is None:
            return data

        _columns = eKonf.ensure_list(self.DATETIME_INFO.get(eKonf.Keys.COLUMNS))
        _format = self.DATETIME_INFO.get(eKonf.Keys.FORMAT, None)
        rcParams = self.DATETIME_INFO.get(eKonf.Keys.rcPARAMS) or {}
        if _columns is None:
            log.info("No datetime column found")
            return data
        if isinstance(data, pd.DataFrame):
            for _col in _columns:
                if _col in data.columns:
                    data[_col] = pd.to_datetime(data[_col], format=_format, **rcParams)
                    log.info(f"converted datetime column {_col}")
        return data

    def append_id(self, _id):
        log.info(f"Adding id [{_id}] to {self.IDs}")
        if self.IDs is None:
            self.IDs = [_id]
        else:
            if isinstance(self.IDs, str):
                self.IDs = [self.IDs]
            self.IDs += [_id]
        log.info(f"Added id [{_id}], now {self.IDs}")

    def append_dataset(self, data, _dataset):
        if _dataset is None:
            return data

        if isinstance(data, pd.DataFrame):
            data[self.DATASET] = _dataset
            if self.DATASET not in self.IDs:
                self.append_id(self.DATASET)
            if self.DATA and self.DATASET not in self.DATA:
                self.DATATYPEs[self.DATASET] = "str"

            log.info(f"Added a column [{self.DATASET}] with value [{_dataset}]")

        return data

    def append_split(self, data, _split):
        if _split is None:
            return data

        if isinstance(data, pd.DataFrame):
            data[self.SPLIT] = _split
            if self.SPLIT not in self.IDs:
                self.append_id(self.SPLIT)
            if self.DATA and self.SPLIT not in self.DATA:
                self.DATATYPEs[self.SPLIT] = "str"

            log.info(f"Added a column [{self.SPLIT}] with value [{_split}]")

        return data

    def set_dtypes(self, data):
        if isinstance(data, pd.DataFrame):
            dtypes = data.dtypes.apply(lambda x: x.name).to_dict()
            self.DATATYPEs = dtypes
        return data

    @property
    def _ID(self):
        return self.Keys._ID.value

    @property
    def ID_SEPARATOR(self):
        return eKonf.Defaults.ID_SEP.value

    @property
    def INFO(self):
        return self.args

    @property
    def DATETIME_INFO(self):
        return self.INFO.get(self.Keys.DATETIME)

    @DATETIME_INFO.setter
    def DATETIME_INFO(self, value):
        self.INFO[eKonf.Keys.DATETIME.value] = value

    @property
    def DATATYPEs(self):
        return self.INFO.get(eKonf.Keys.DATA)

    @DATATYPEs.setter
    def DATATYPEs(self, value):
        self.INFO[eKonf.Keys.DATA.value] = value

    @property
    def COLUMNs(self):
        return self.INFO.get(eKonf.Keys.COLUMNS) or {}

    @COLUMNs.setter
    def COLUMNs(self, value):
        self.INFO[eKonf.Keys.COLUMNS.value] = value

    @property
    def DATA(self):
        if self.DATATYPEs is None:
            return None
        return list(self.DATATYPEs.keys())

    @property
    def DATASET(self):
        return eKonf.Keys.DATASET.value

    @property
    def INDEX(self):
        return self.COLUMNs.get(eKonf.Keys.INDEX) or eKonf.Keys.INDEX.value

    @INDEX.setter
    def INDEX(self, value):
        self.COLUMNs[eKonf.Keys.INDEX.value] = value

    @property
    def ID(self):
        return eKonf.Keys.ID.value

    @property
    def IDs(self):
        return eKonf.ensure_list(self.COLUMNs.get(self.ID))

    @IDs.setter
    def IDs(self, value):
        self.COLUMNs[self.ID] = value

    @property
    def SPLIT(self):
        return eKonf.Keys.SPLIT.value


class CorpusInfo(BaseInfo):
    def __init__(self, **args):
        super().__init__(**args)

    def to_timestamp(self, data, metadata=None):
        if self.TIMESTAMP_INFO is None:
            return data, metadata

        _key = self.TIMESTAMP_INFO.get(eKonf.Keys.KEY)
        _format = self.TIMESTAMP_INFO.get(eKonf.Keys.FORMAT)
        rcParams = self.TIMESTAMP_INFO.get(eKonf.Keys.rcPARAMS) or {}
        if _key is None:
            log.info("No timestamp key found")
            return data, metadata
        if isinstance(data, pd.DataFrame):
            if _key in data.columns:
                data[self.TIMESTAMP] = pd.to_datetime(
                    data[_key], format=_format, **rcParams
                )
                log.info(f"Loaded timestamp column {self.TIMESTAMP}")
            elif metadata is not None and _key in metadata.columns:
                metadata[self.TIMESTAMP] = pd.to_datetime(
                    metadata[_key], format=_format, **rcParams
                )
                df_dt = metadata[self.MERGE_META_ON + [self.TIMESTAMP]].copy()
                data = data.merge(df_dt, on=self.MERGE_META_ON, how="left")
                # metadata.drop(self.TIMESTAMP, axis=1, inplace=True)
                log.info(f"Timestamp column {self.TIMESTAMP} added to data")
        return data, metadata

    def combine_texts(self, data):
        if self.TEXTs is None:
            return data

        if isinstance(data, pd.DataFrame):
            data[self.TEXTs] = data[self.TEXTs].fillna("")
            if len(self.TEXTs) > 1:
                data[self.TEXT] = data[self.TEXTs].apply(
                    lambda row: self.SEGMENT_SEP.join(row.values.astype(str)),
                    axis=1,
                )
                self.DATATYPEs = {
                    k: v for k, v in self.DATATYPEs.items() if k not in self.TEXTs
                }
                self.DATATYPEs[self.TEXT] = "str"

        return data

    def merge_metadata(self, data, metadata):
        if metadata is None:
            return data
        meta_cols = [col for col in metadata.columns if col not in data.columns]
        meta_cols += self.MERGE_META_ON
        data = data.merge(metadata[meta_cols], on=self.MERGE_META_ON, how="left")
        return data

    def append_split_to_meta(self, metadata, _split):
        if _split is None:
            return metadata

        if isinstance(metadata, pd.DataFrame):
            metadata[self.SPLIT] = _split
            if self.METADATA and self.SPLIT not in self.METADATA:
                self.METATYPEs[self.SPLIT] = "str"

            log.info(f"Added a column [{self.SPLIT}] with value [{_split}]")

        return metadata

    def append_corpus(self, data, _corpus):
        if _corpus is None:
            return data

        if isinstance(data, pd.DataFrame):
            data[self.CORPUS] = _corpus
            if self.CORPUS not in self.IDs:
                self.append_id(self.CORPUS)
            if self.DATA and self.CORPUS not in self.DATA:
                self.DATATYPEs[self.CORPUS] = "str"
            if self.METADATA and self.CORPUS not in self.METADATA:
                self.METATYPEs[self.CORPUS] = "str"

            log.info(f"Added a column [{self.CORPUS}] with value [{_corpus}]")

        return data

    @property
    def MERGE_META_ON(self):
        return eKonf.ensure_list(self.COLUMNs.get(eKonf.Keys.META_MERGE_ON)) or self.IDs

    @MERGE_META_ON.setter
    def MERGE_META_ON(self, value):
        self.COLUMNs[eKonf.Keys.META_MERGE_ON.value] = value

    @property
    def TEXT(self):
        return eKonf.Keys.TEXT.value

    @property
    def TEXTs(self):
        return eKonf.ensure_list(self.COLUMNs.get(self.TEXT))

    @TEXTs.setter
    def TEXTs(self, value):
        self.COLUMNs[self.TEXT] = value

    @property
    def METADATA(self):
        if self.METATYPEs is None:
            return None
        return list(self.METATYPEs.keys())

    @property
    def TIMESTAMP(self):
        return eKonf.Keys.TIMESTAMP.value

    @property
    def CORPUS(self):
        return eKonf.Keys.CORPUS.value

    @property
    def METATYPEs(self):
        return self.INFO.get(eKonf.Keys.META)

    @METATYPEs.setter
    def METATYPEs(self, value):
        self.INFO[eKonf.Keys.META.value] = value

    @property
    def TIMESTAMP_INFO(self):
        return self.INFO.get(self.TIMESTAMP)

    @TIMESTAMP_INFO.setter
    def TIMESTAMP_INFO(self, value):
        self.INFO[self.TIMESTAMP] = value

    @property
    def SEGMENT_SEP(self):
        return codecs.decode(
            self.INFO.get("segment_separator", "\n\n"), "unicode_escape"
        )

    @property
    def SENTENCE_SEP(self):
        return codecs.decode(
            self.INFO.get("sentence_separator", "\n"), "unicode_escape"
        )


class DatasetInfo(BaseInfo):
    def __init__(self, **args):
        super().__init__(**args)


class FeatureInfo(BaseInfo):
    def __init__(self, **args):
        super().__init__(**args)

    @property
    def Y(self):
        return self.COLUMNs.get(eKonf.Keys.Y)

    @Y.setter
    def Y(self, value):
        self.COLUMNs[eKonf.Keys.Y.value] = value

    @property
    def X(self):
        return eKonf.ensure_list(self.COLUMNs.get(eKonf.Keys.X))

    @X.setter
    def X(self, value):
        self.COLUMNs[eKonf.Keys.X.value] = value



from ..types import Config


class ExponentialLRSchedulerConfig(Config):
    """Settings for Exponential decay learning rate scheduler."""

    gamma: float = 0.9


class StepLRSchedulerConfig(Config):
    """Settings for Step decay learning rate scheduler."""

    gamma: float = 0.9
    step_size: int = 5


class LRSchedulerConfig(Config):
    """Settings common to all LRSchedulers.

    Parameters
    ----------
    warmup_steps
        The number of warmup_units over which to perform learning rate warmup (a linear
        increase from 0 to the specified lr)
    warmup_unit
        The unit to use when counting warmup (one of ["batches", "epochs"])
    warmup_percentage
        The percentage of the training procedure to warm up over (ignored if
        warmup_steps is non-zero)
    min_lr
        The minimum learning rate to use during training (the learning rate specified
        by a learning rate scheduler will be rounded up to this if it is lower)
    exponential_config
        Extra settings for the ExponentialLRScheduler
    step_config
        Extra settings for the StepLRScheduler
    """

    warmup_steps: float = 0  # warm up steps
    warmup_unit: str = "batches"  # [epochs, batches]
    warmup_percentage: float = 0.0  # warm up percentage
    min_lr: float = 0.0  # minimum learning rate
    exponential_config: ExponentialLRSchedulerConfig = (
        ExponentialLRSchedulerConfig()  # type:ignore
    )
    step_config: StepLRSchedulerConfig = StepLRSchedulerConfig()  # type:ignore



import os
import subprocess
import pandas as pd
from collections import namedtuple
from pathlib import Path
import logging


log = logging.getLogger(__name__)


Feature = namedtuple(
    "Feature",
    [
        "pos",
        "semantic",
        "has_jongseong",
        "reading",
        "type",
        "start_pos",
        "end_pos",
        "expression",
    ],
)
# ('합니다',
#   Feature(pos='XSA+EF', semantic=None, has_jongseong=False, reading='합니다', type='Inflect',
#           start_pos='XSA', end_pos='EF',
#           expression='하/XSA/*+ᄇ니다/EF/*')),


def _extract_feature(values):
    # Reference:
    # - http://taku910.github.io/mecab/learn.html
    # - https://docs.google.com/spreadsheets/d/1-9blXKjtjeKZqsf4NzHeYJCrr49-nXeRF6D80udfcwY
    # - https://bitbucket.org/eunjeon/mecab-ko-dic/src/master/utils/dictionary/lexicon.py

    # feature = <pos>,<semantic>,<has_jongseong>,<reading>,<type>,<start_pos>,<end_pos>,<expression>
    assert len(values) == 8

    values = [value if value != "*" else None for value in values]
    feature = dict(zip(Feature._fields, values))
    feature["has_jongseong"] = {"T": True, "F": False}.get(feature["has_jongseong"])

    return Feature(**feature)


class MeCabError(Exception):
    pass


class MeCab:  # APIs are inspried by KoNLPy
    def __init__(
        self,
        dicdir=None,
        userdic_path=None,
        backend="mecab-python3",
        verbose=False,
        **kwargs,
    ):
        import mecab_ko_dic

        self.dicdir = None
        self.userdic_path = None
        self.verbose = verbose

        DICDIR = mecab_ko_dic.DICDIR
        mecabrc = os.path.join(DICDIR, "mecabrc")

        self.backend = backend
        assert self.backend in [
            "fugashi",
            "mecab-python3",
        ], "Wrong backend! Currently, we support [`fugashi`, `mecab-python3`] backend."
        if self.verbose:
            log.info(f"MeCab uses {self.backend} as backend.")

        if not dicdir:
            dicdir = DICDIR
        self.dicdir = dicdir
        MECAB_ARGS = '-r "{}" -d "{}" '.format(mecabrc, dicdir)
        if userdic_path:
            self.userdic_path = userdic_path
            MECAB_ARGS += '-u "{}" '.format(userdic_path)
        if self.verbose:
            log.info(
                f"Mecab uses system dictionary: {dicdir}, user dictionary: {userdic_path}"
            )
        try:
            if self.backend == "fugashi":
                try:
                    import fugashi as _mecab
                except ImportError:
                    raise ImportError(
                        "\n"
                        "You must install `fugashi` if you want to use `fugashi` backend.\n"
                    )
                self.tagger = _mecab.GenericTagger(MECAB_ARGS)
                self.dictionary_info = self.tagger.dictionary_info
                self.sysdic_path = self.dictionary_info[0]["filename"]
            else:
                try:
                    import MeCab as _mecab
                except ImportError:
                    raise ImportError(
                        "\n"
                        "You must install `mecab-python3` if you want to use `mecab-python3` backend.\n"
                    )
                self.tagger = _mecab.Tagger(MECAB_ARGS)
                self.version = _mecab.VERSION
        except RuntimeError:
            raise Exception(
                'The MeCab dictionary does not exist at "%s". Is the dictionary correctly installed?\nYou can also try entering the dictionary path when initializing the MeCab class: "MeCab(\'/some/dic/path\')"'
                % self.dicdir
            )
        except NameError:
            raise Exception("Check if MeCab is installed correctlly.")

    def _parse_fugashi(self, text):
        return [
            (node.surface, _extract_feature(node.feature)) for node in self.tagger(text)
        ]

    def _parse_mecab(self, text):
        m = self.tagger.parseToNode(text)
        nodes = []
        while m:
            feature = m.feature.split(",")
            if "BOS/EOS" != feature[0]:
                nodes.append((m.surface, _extract_feature(feature)))
            m = m.next
        return nodes

    def parse(self, text):
        if self.backend == "fugashi":
            return self._parse_fugashi(text)
        else:
            return self._parse_mecab(text)

    def pos(
        self,
        sentence,
        flatten=True,
        concat_surface_and_pos=False,
        include_whitespace_token=True,
    ):
        sentence = sentence.strip()
        if include_whitespace_token:
            sent_ptr = 0
            res = []

            for token, pos in self._pos(sentence, flatten=flatten):
                if sentence[sent_ptr] == " ":
                    # Move pointer to whitespace token to reserve whitespace
                    # cf. to prevent double white-space, move pointer to next eojeol
                    while sentence[sent_ptr] == " ":
                        sent_ptr += 1
                    res.append((" ", "SP"))
                res.append((token, pos))
                sent_ptr += len(token)
        else:
            res = self._pos(sentence, flatten=flatten)

        return [s[0] + "/" + s[1] if concat_surface_and_pos else s for s in res]

    def _pos(self, sentence, flatten=True):
        if flatten:
            return [(surface, feature.pos) for surface, feature in self.parse(sentence)]
        else:
            res = []
            for surface, feature in self.parse(sentence):
                if feature.expression is None:
                    res.append((surface, feature.pos))
                else:
                    for elem in feature.expression.split("+"):
                        s = elem.split("/")
                        res.append((s[0], s[1]))
            return res

    def morphs(self, sentence, flatten=True, include_whitespace_token=True):
        return [
            surface
            for surface, _ in self.pos(
                sentence,
                flatten=flatten,
                concat_surface_and_pos=False,
                include_whitespace_token=include_whitespace_token,
            )
        ]

    def nouns(
        self,
        sentence,
        flatten=True,
        include_whitespace_token=True,
        noun_pos=["NNG", "NNP", "XSN", "SL", "XR", "NNB", "NR"],
    ):
        return [
            surface
            for surface, pos in self.pos(
                sentence,
                flatten=flatten,
                concat_surface_and_pos=False,
                include_whitespace_token=include_whitespace_token,
            )
            if pos in noun_pos
        ]


DicEntry = namedtuple(
    "DicEntry",
    [
        "surface",
        "left_id",
        "right_id",
        "cost",
        "pos",
        "semantic",
        "has_jongseong",
        "reading",
        "type",
        "start_pos",
        "end_pos",
        "expression",
    ],
    defaults=[None, None, None, None, "NNP", "*", "T", None, "*", "*", "*", "*"],
)

ContextEntry = namedtuple(
    "ContextEntry",
    [
        "id",
        "pos",
        "semantic",
        "has_jongseong",
        "reading",
        "type",
        "start_pos",
        "end_pos",
        "expression",
    ],
    defaults=[None, "*", "*", "*", "*", "*", "*", "*"],
)


def iternamedtuples(df):
    Row = namedtuple("Row", df.columns)
    for row in df.itertuples():
        yield Row(*row[1:])


def has_jongseong(c):
    return int((ord(c[-1]) - 0xAC00) % 28) != 0


class MecabDicConfig:
    def __init__(self, userdic_path=None):
        import mecab_ko_dic

        if userdic_path:
            self.load_userdic(userdic_path)
        else:
            self.userdic = {}
        self.dicdir = mecab_ko_dic.DICDIR
        self.left_ids = self.load_context_ids("left-id.def")
        self.right_ids = self.load_context_ids("right-id.def")

    def load_context_ids(self, id_file):
        id_file = os.path.join(self.dicdir, id_file)
        context_ids = []
        with open(id_file, "r", encoding="utf-8") as f:
            for line in f:
                id, vals = line.split()
                entry = ContextEntry(id, *vals.split(","))
                context_ids.append(entry)
        return context_ids

    def find_left_context_id(self, search):
        for entry in self.left_ids:
            if entry.pos == search.pos and entry.semantic == search.semantic:
                return entry.id

    def find_right_context_id(self, search):
        for entry in self.right_ids:
            if (
                entry.pos == search.pos
                and entry.semantic == search.semantic
                and entry.has_jongseong == search.has_jongseong
            ):
                return entry.id

    def load_userdic(self, userdic_path):
        userdic_path = Path(userdic_path)

        if userdic_path.is_dir():
            self.userdic = {}
            for f in userdic_path.glob("*.csv"):
                df = pd.read_csv(f, names=DicEntry._fields)
                dic = {e.surface: e for e in iternamedtuples(df)}
                self.userdic = {**self.userdic, **dic}
        else:
            df = pd.read_csv(userdic_path, names=DicEntry._fields)
            self.userdic = {e.surface: e for e in iternamedtuples(df)}
        print(" No. of user dictionary entires loaded: %d" % len(self.userdic))

    def add_entry_to_userdic(
        self, surface, pos="NNP", semantic="*", reading=None, cost=1000
    ):
        entry = DicEntry(
            surface=surface,
            cost=cost,
            pos=pos,
            semantic=semantic,
            has_jongseong={True: "T", False: "F"}.get(has_jongseong(surface)),
            reading=surface if reading is None else reading,
        )
        entry = entry._replace(
            left_id=self.find_left_context_id(entry),
            right_id=self.find_right_context_id(entry),
        )
        self.userdic[surface] = entry

    def adjust_context_ids(self):
        for entry in self.userdic.values():
            entry = entry._replace(
                left_id=self.find_left_context_id(entry),
                right_id=self.find_right_context_id(entry),
            )
            self.userdic[entry.surface] = entry

    def adjust_costs(self, cost=1000):
        for surface, entry in self.userdic.items():
            self.userdic[surface] = entry._replace(cost=cost)

    def save_userdic(self, save_path):
        if len(self.userdic) > 0:
            df = pd.DataFrame(self.userdic.values())
            df.to_csv(save_path, header=False, index=False)
            self.userdic_path = save_path
            print("Saved the userdic to {}".format(save_path))
        else:
            print("No userdic to save...")

    def build_userdic(self, built_userdic_path, userdic_path=None):
        if userdic_path:
            self.userdic_path = userdic_path
        args = '-d "{}" -u "{}" {}'.format(
            self.dicdir, built_userdic_path, self.userdic_path
        )
        # print(args)
        subprocess.run(["fugashi-build-dict", args])



from functools import partial
from typing import Union

import numpy as np
import pandas as pd
from dask import dataframe as dd
from dask.distributed import Client

from .core import BaseLFApplier, _FunctionCaller
from .pandas import apply_lfs_to_data_point, rows_to_triplets

Scheduler = Union[str, Client]


class DaskLFApplier(BaseLFApplier):
    """LF applier for a Dask DataFrame.

    Dask DataFrames consist of partitions, each being a Pandas DataFrame.
    This allows for efficient parallel computation over DataFrame rows.
    For more information, see https://docs.dask.org/en/stable/dataframe.html
    """

    def apply(
        self,
        df: dd.DataFrame,
        scheduler: Scheduler = "processes",
        fault_tolerant: bool = False,
    ) -> np.ndarray:
        """Label Dask DataFrame of data points with LFs.

        Parameters
        ----------
        df
            Dask DataFrame containing data points to be labeled by LFs
        scheduler
            A Dask scheduling configuration: either a string option or
            a ``Client``. For more information, see
            https://docs.dask.org/en/stable/scheduling.html#
        fault_tolerant
            Output ``-1`` if LF execution fails?

        Returns
        -------
        np.ndarray
            Matrix of labels emitted by LFs
        """
        f_caller = _FunctionCaller(fault_tolerant)
        apply_fn = partial(apply_lfs_to_data_point, lfs=self._lfs, f_caller=f_caller)
        map_fn = df.map_partitions(lambda p_df: p_df.apply(apply_fn, axis=1))
        labels = map_fn.compute(scheduler=scheduler)
        labels_with_index = rows_to_triplets(labels)
        return self._numpy_from_row_data(labels_with_index)


class PandasParallelLFApplier(DaskLFApplier):
    """Parallel LF applier for a Pandas DataFrame.

    Creates a Dask DataFrame from a Pandas DataFrame, then uses
    ``DaskLFApplier`` to label data in parallel. See ``DaskLFApplier``.
    """

    def apply(  # type: ignore
        self,
        df: pd.DataFrame,
        n_parallel: int = 2,
        scheduler: Scheduler = "processes",
        fault_tolerant: bool = False,
    ) -> np.ndarray:
        """Label Pandas DataFrame of data points with LFs in parallel using Dask.

        Parameters
        ----------
        df
            Pandas DataFrame containing data points to be labeled by LFs
        n_parallel
            Parallelism level for LF application. Corresponds to ``npartitions``
            in constructed Dask DataFrame. For ``scheduler="processes"``, number
            of processes launched. Recommended to be no more than the number
            of cores on the running machine.
        scheduler
            A Dask scheduling configuration: either a string option or
            a ``Client``. For more information, see
            https://docs.dask.org/en/stable/scheduling.html#
        fault_tolerant
            Output ``-1`` if LF execution fails?

        Returns
        -------
        np.ndarray
            Matrix of labels emitted by LFs
        """
        if n_parallel < 2:
            raise ValueError(
                "n_parallel should be >= 2. "
                "For single process Pandas, use PandasLFApplier."
            )
        df = dd.from_pandas(df, npartitions=n_parallel)
        return super().apply(df, scheduler=scheduler, fault_tolerant=fault_tolerant)



import logging
from ekorpkit.io.load.list import load_wordlist
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class Stopwords:
    def __init__(self, **kwargs):
        kwargs = eKonf.to_dict(kwargs)

        self.name = kwargs.get("name", "stopwords")
        self._lowercase = kwargs.get("lowercase", False)
        self.verbose = kwargs.get("verbose", False)

        self._stopwords_path = kwargs.get("stopwords_path", None)
        if self._stopwords_path is not None:
            self._stopwords_list = load_wordlist(
                self._stopwords_path, lowercase=self._lowercase, verbose=self.verbose
            )
            if self.verbose:
                log.info(
                    f"Loaded {len(self._stopwords_list)} stopwords from {self._stopwords_path}"
                )
        else:
            self._stopwords_list = []

        self._stopwords_fn = lambda x: False
        stopwords = kwargs.get("stopwords")

        if callable(stopwords):
            self._stopwords_fn = stopwords
            if self.verbose:
                log.info(f"Using custom stopwords function: {str(self._stopwords_fn)}")
        elif isinstance(stopwords, list):
            if self._lowercase:
                stopwords = [w.lower() for w in stopwords]
            if self.verbose:
                log.info(f"Loaded {len(stopwords)} stopwords from list")
            self._stopwords_list += stopwords

        nltk_stopwords = kwargs.get("nltk_stopwords")
        if nltk_stopwords:
            self._stopwords_list += self._load_nltk_stopwords(nltk_stopwords)
            if self.verbose:
                log.info(f"Loaded {len(self._stopwords_list)} stopwords in total")

    def __call__(self, word):
        """Calling a stopwords instance like a function just calls the is_stopword method."""
        return self.is_stopword(word)

    def is_stopword(self, word):
        """
        :type word: str
        :returns: bool
        """
        _word = word.lower() if self._lowercase else word
        return self._stopwords_fn(_word) or _word in self._stopwords_list

    def _load_nltk_stopwords(self, language="english"):
        """
        :type language: str
        :returns: list
        """
        import nltk
        from nltk.corpus import stopwords

        nltk.download("stopwords", quiet=True)
        if language in stopwords.fileids():
            if self.verbose:
                log.info(f"Loading stopwords for {language} from NLTK")
            return stopwords.words(language)

        log.warning(f"No stopwords for {language} in NLTK")
        return []



import os
import logging
from ekorpkit import eKonf
from .base import BaseSet


log = logging.getLogger(__name__)


class Dataset(BaseSet):
    """Dataset class."""

    def __init__(self, **args):
        super().__init__(**args)
        if isinstance(self.name, list):
            self.name = self.name[0]
        use_name_as_subdir = args.get("use_name_as_subdir", True)
        if use_name_as_subdir:
            self.data_dir = os.path.join(self.data_dir, self.name)

        self.load_info()
        self.load_column_info()

        if self.auto.build:
            if self.force.build or not eKonf.exists(
                self.data_dir, self.data_files[self.SPLITS.TRAIN]
            ):
                self.build()
        if self.auto.load:
            self.load()

    @property
    def train_data(self):
        if self.SPLITS.TRAIN not in self.splits:
            return None
        return self.splits[self.SPLITS.TRAIN]

    @train_data.setter
    def train_data(self, data):
        self.splits[self.SPLITS.TRAIN.value] = data

    @property
    def dev_data(self):
        if self.SPLITS.DEV not in self.splits:
            return None
        return self.splits[self.SPLITS.DEV]

    @dev_data.setter
    def dev_data(self, data):
        self.splits[self.SPLITS.DEV.value] = data

    @property
    def test_data(self):
        if self.SPLITS.TEST not in self.splits:
            return None
        return self.splits[self.SPLITS.TEST]

    @test_data.setter
    def test_data(self, data):
        self.splits[self.SPLITS.TEST.value] = data



import logging
import torch
from transformers import pipeline
from huggingface_hub import InferenceApi
from ekorpkit.utils.func import elapsed_timer
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class BloomDemo:
    def __init__(
        self,
        model_uri="bigscience/bloom",
        device=0,
        hf_user_access_token=None,
        **kwargs,
    ):
        if kwargs:
            self.args = eKonf.to_config(kwargs)
        else:
            self.args = {}
        self.model_uri = model_uri
        self.device = device
        self.hf_user_access_token = hf_user_access_token
        self.auto_load = self.args.get("auto", {}).get("load", False)

        self.inference_api = None
        self.pipe = None
        self.TRANSFORMERS_CACHE = eKonf.osenv("TRANSFORMERS_CACHE")

    def init_pipe(self, model_uri=None, task="text-generation", device=None):
        if model_uri is None:
            model_uri = self.model_uri
        if device is None:
            device = self.device if torch.cuda.is_available() else -1
        with elapsed_timer(format_time=True) as elapsed:
            self.pipe = pipeline(
                model=model_uri,
                task=task,
                device=device,
                torch_dtype=torch.bfloat16,
                cache_dir=self.TRANSFORMERS_CACHE,
            )
            print(" >> elapsed time to load model: {}".format(elapsed()))

    def init_api(self, model_uri=None, token=None):
        if model_uri is None:
            model_uri = self.model_uri
        if token is None:
            token = self.hf_user_access_token or eKonf.osenv("HF_USER_ACCESS_TOKEN")
        self.inference_api = InferenceApi(model_uri, token=token)
        print(" >> Inference API initialized")

    def infer(
        self,
        prompt,
        inference_method="api",
        temperature=0.7,
        top_k=None,
        top_p=None,
        max_new_tokens=50,
        repetition_penalty=None,
        do_sample=False,
        num_return_sequences=1,
        num_beams=None,
        no_repeat_ngram_size=None,
        early_stopping=False,
        return_full_text=True,
        **kwargs,
    ):
        if inference_method == "api":
            response = self.infer_api(
                prompt,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
                max_new_tokens=max_new_tokens,
                repetition_penalty=repetition_penalty,
                do_sample=do_sample,
                num_return_sequences=num_return_sequences,
                num_beams=num_beams,
                no_repeat_ngram_size=no_repeat_ngram_size,
                early_stopping=early_stopping,
                return_full_text=return_full_text,
                **kwargs,
            )
        else:
            response = self.infer_local(
                prompt,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
                max_new_tokens=max_new_tokens,
                repetition_penalty=repetition_penalty,
                do_sample=do_sample,
                num_return_sequences=num_return_sequences,
                num_beams=num_beams,
                no_repeat_ngram_size=no_repeat_ngram_size,
                early_stopping=early_stopping,
                return_full_text=return_full_text,
                **kwargs,
            )

        generated_text = response[0]["generated_text"]
        return generated_text

    def infer_api(
        self,
        prompt,
        temperature=0.7,
        top_k=None,
        top_p=None,
        max_new_tokens=50,
        repetition_penalty=None,
        do_sample=False,
        num_return_sequences=1,
        num_beams=None,
        no_repeat_ngram_size=None,
        early_stopping=False,
        return_full_text=True,
        seed=123,
        **kwargs,
    ):
        if self.inference_api is None:
            self.init_api()

        top_k = None if top_k == 0 else top_k
        top_p = None if num_beams else top_p
        num_beams = None if num_beams == 0 else num_beams
        no_repeat_ngram_size = None if num_beams is None else no_repeat_ngram_size
        early_stopping = None if num_beams is None else num_beams > 0
        max_new_tokens = min(int(max_new_tokens), 2047)

        params = {
            "max_new_tokens": max_new_tokens,
            "top_k": top_k,
            "top_p": top_p,
            "temperature": temperature,
            "do_sample": do_sample,
            "early_stopping": early_stopping,
            "no_repeat_ngram_size": no_repeat_ngram_size,
            "num_beams": num_beams,
            "return_full_text": return_full_text,
            "repetition_penalty": repetition_penalty,
            "seed": seed,
        }

        response = self.inference_api(prompt, params=params)
        return response

    def infer_local(
        self,
        prompt,
        temperature=0.7,
        top_k=None,
        top_p=None,
        max_new_tokens=50,
        repetition_penalty=None,
        do_sample=False,
        num_return_sequences=1,
        num_beams=None,
        no_repeat_ngram_size=None,
        early_stopping=False,
        return_full_text=True,
        **kwargs,
    ):
        if self.pipe is None:
            self.init_pipe()
        max_new_tokens = min(int(max_new_tokens), 2047)

        response = self.pipe(
            prompt,
            temperature=temperature,  # 0 to 1
            top_k=top_k,
            top_p=top_p,  # None, 0-1
            max_new_tokens=max_new_tokens,  # up to 2047 theoretically
            return_full_text=return_full_text,  # include prompt or not.
            repetition_penalty=repetition_penalty,  # None, 0-100 (penalty for repeat tokens.
            do_sample=do_sample,  # True: use sampling, False: Greedy decoding.
            num_return_sequences=num_return_sequences,
            num_beams=num_beams,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
        )
        return response



import logging
import os
import shutil
import tarfile
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Tuple
from urllib.parse import urlparse
from zipfile import ZipFile, is_zipfile

from .cache_file import CacheFile
from .common import PathOrStr, get_cache_dir
from .file_lock import FileLock
from .meta import Meta
from .schemes import get_scheme_client, get_supported_schemes, hf_get_from_cache
from .util import (
    _lock_file_path,
    _meta_file_path,
    check_tarfile,
    find_latest_cached,
    resource_to_filename,
)

if TYPE_CHECKING:
    from rich.progress import Progress

logger = logging.getLogger("cached_path")


def cached_path(
    url_or_filename: PathOrStr,
    cache_dir: Optional[PathOrStr] = None,
    extract_archive: bool = False,
    force_extract: bool = False,
    quiet: bool = False,
    progress: Optional["Progress"] = None,
) -> Path:
    """
    Given something that might be a URL or local path, determine which.
    If it's a remote resource, download the file and cache it, and
    then return the path to the cached file. If it's already a local path,
    make sure the file exists and return the path.

    For URLs, the following schemes are all supported out-of-the-box:

    * ``http`` and ``https``,
    * ``s3`` for objects on `AWS S3`_,
    * ``gs`` for objects on `Google Cloud Storage (GCS)`_, and
    * ``hf`` for objects or repositories on `HuggingFace Hub`_.

    You can also extend ``cached_path()`` to handle more schemes with :func:`add_scheme_client()`.

    .. _AWS S3: https://aws.amazon.com/s3/
    .. _Google Cloud Storage (GCS): https://cloud.google.com/storage
    .. _HuggingFace Hub: https://huggingface.co/

    Examples
    --------

    To download a file over ``https``::

        cached_path("https://github.com/allenai/cached_path/blob/main/README.md")

    To download an object on GCS::

        cached_path("gs://allennlp-public-models/lerc-2020-11-18.tar.gz")

    To download the PyTorch weights for the model `epwalsh/bert-xsmall-dummy`_
    on HuggingFace, you could do::

        cached_path("hf://epwalsh/bert-xsmall-dummy/pytorch_model.bin")

    For paths or URLs that point to a tarfile or zipfile, you can append the path
    to a specific file within the archive to the ``url_or_filename``, preceeded by a "!".
    The archive will be automatically extracted (provided you set ``extract_archive`` to ``True``),
    returning the local path to the specific file. For example::

        cached_path("model.tar.gz!weights.th", extract_archive=True)

    .. _epwalsh/bert-xsmall-dummy: https://huggingface.co/epwalsh/bert-xsmall-dummy

    Parameters
    ----------

    url_or_filename :
        A URL or path to parse and possibly download.

    cache_dir :
        The directory to cache downloads. If not specified, the global default cache directory
        will be used (``~/.cache/cached_path``). This can be set to something else with
        :func:`set_cache_dir()`.

    extract_archive :
        If ``True``, then zip or tar.gz archives will be automatically extracted.
        In which case the directory is returned.

    force_extract :
        If ``True`` and the file is an archive file, it will be extracted regardless
        of whether or not the extracted directory already exists.

        .. caution::
            Use this flag with caution! This can lead to race conditions if used
            from multiple processes on the same file.

    quiet :
        If ``True``, progress displays won't be printed.

    progress :
        A custom progress display to use. If not set and ``quiet=False``, a default display
        from :func:`~cached_path.get_download_progress()` will be used.

    Returns
    -------
    :class:`pathlib.Path`
        The local path to the (potentially cached) resource.

    Raises
    ------
    ``FileNotFoundError``

        If the resource cannot be found locally or remotely.

    ``ValueError``
        When the URL is invalid.

    ``Other errors``
        Other error types are possible as well depending on the client used to fetch
        the resource.

    """
    cache_dir = Path(cache_dir if cache_dir else get_cache_dir()).expanduser()
    cache_dir.mkdir(parents=True, exist_ok=True)

    if not isinstance(url_or_filename, str):
        url_or_filename = str(url_or_filename)

    file_path: Path
    extraction_path: Optional[Path] = None
    etag: Optional[str] = None

    # If we're using the /a/b/foo.zip!c/d/file.txt syntax, handle it here.
    exclamation_index = url_or_filename.find("!")
    if extract_archive and exclamation_index >= 0:
        archive_path = url_or_filename[:exclamation_index]
        file_name = url_or_filename[exclamation_index + 1 :]

        # Call 'cached_path' recursively now to get the local path to the archive itself.
        cached_archive_path = cached_path(
            archive_path,
            cache_dir=cache_dir,
            extract_archive=True,
            force_extract=force_extract,
            quiet=quiet,
            progress=progress,
        )
        if not cached_archive_path.is_dir():
            raise ValueError(
                f"{url_or_filename} uses the ! syntax, but does not specify an archive file."
            )

        # Now return the full path to the desired file within the extracted archive,
        # provided it exists.
        file_path = cached_archive_path / file_name
        if not file_path.exists():
            raise FileNotFoundError(f"'{file_name}' not found within '{archive_path}'")

        return file_path

    parsed = urlparse(url_or_filename)

    if parsed.scheme in get_supported_schemes():
        # URL, so get it from the cache (downloading if necessary)
        file_path, etag = get_from_cache(url_or_filename, cache_dir, quiet=quiet, progress=progress)

        if extract_archive and (is_zipfile(file_path) or tarfile.is_tarfile(file_path)):
            # This is the path the file should be extracted to.
            # For example ~/.cached_path/cache/234234.21341 -> ~/.cached_path/cache/234234.21341-extracted
            extraction_path = file_path.parent / (file_path.name + "-extracted")

    else:
        url_or_filename = Path(url_or_filename).expanduser()

        if url_or_filename.exists():
            # File, and it exists.
            file_path = url_or_filename
            # Normalize the path.
            url_or_filename = url_or_filename.resolve()

            if (
                extract_archive
                and file_path.is_file()
                and (is_zipfile(file_path) or tarfile.is_tarfile(file_path))
            ):
                # We'll use a unique directory within the cache to root to extract the archive to.
                # The name of the directory is a hash of the resource file path and it's modification
                # time. That way, if the file changes, we'll know when to extract it again.
                extraction_name = (
                    resource_to_filename(url_or_filename, str(os.path.getmtime(file_path)))
                    + "-extracted"
                )
                extraction_path = cache_dir / extraction_name

        elif parsed.scheme == "":
            # File, but it doesn't exist.
            raise FileNotFoundError(f"file {url_or_filename} not found")

        else:
            # Something unknown
            raise ValueError(f"unable to parse {url_or_filename} as a URL or as a local path")

    if extraction_path is not None:
        # If the extracted directory already exists (and is non-empty), then no
        # need to create a lock file and extract again unless `force_extract=True`.
        if os.path.isdir(extraction_path) and os.listdir(extraction_path) and not force_extract:
            return extraction_path

        # Extract it.
        with FileLock(_lock_file_path(extraction_path)):
            # Check again if the directory exists now that we've acquired the lock.
            if os.path.isdir(extraction_path) and os.listdir(extraction_path):
                if force_extract:
                    logger.warning(
                        "Extraction directory for %s (%s) already exists, "
                        "overwriting it since 'force_extract' is 'True'",
                        url_or_filename,
                        extraction_path,
                    )
                else:
                    return extraction_path

            logger.info("Extracting %s to %s", url_or_filename, extraction_path)
            shutil.rmtree(extraction_path, ignore_errors=True)

            # We extract first to a temporary directory in case something goes wrong
            # during the extraction process so we don't end up with a corrupted cache.
            tmp_extraction_dir = tempfile.mkdtemp(dir=os.path.split(extraction_path)[0])
            try:
                if is_zipfile(file_path):
                    with ZipFile(file_path, "r") as zip_file:
                        zip_file.extractall(tmp_extraction_dir)
                        zip_file.close()
                else:
                    tar_file = tarfile.open(file_path)
                    check_tarfile(tar_file)
                    tar_file.extractall(tmp_extraction_dir)
                    tar_file.close()
                # Extraction was successful, rename temp directory to final
                # cache directory and dump the meta data.
                os.replace(tmp_extraction_dir, extraction_path)
                meta = Meta.new(
                    url_or_filename,
                    extraction_path,
                    etag=etag,
                    extraction_dir=True,
                )
                meta.to_file()
            finally:
                shutil.rmtree(tmp_extraction_dir, ignore_errors=True)

        return extraction_path

    return file_path


def get_from_cache(
    url: str,
    cache_dir: Optional[PathOrStr] = None,
    quiet: bool = False,
    progress: Optional["Progress"] = None,
) -> Tuple[Path, Optional[str]]:
    """
    Given a URL, look for the corresponding dataset in the local cache.
    If it's not there, download it. Then return the path to the cached file and the ETag.
    """
    cache_dir = Path(cache_dir if cache_dir else get_cache_dir())

    if url.startswith("hf://"):
        return hf_get_from_cache(url, cache_dir), None

    client = get_scheme_client(url)

    # Get eTag to add to filename, if it exists.
    try:
        etag = client.get_etag()
    except client.recoverable_errors:  # type: ignore
        # We might be offline, in which case we don't want to throw an error
        # just yet. Instead, we'll try to use the latest cached version of the
        # target resource, if it exists. We'll only throw an exception if we
        # haven't cached the resource at all yet.
        logger.warning(
            "Connection error occurred while trying to fetch ETag for %s. "
            "Will attempt to use latest cached version of resource",
            url,
        )
        latest_cached = find_latest_cached(url, cache_dir)
        if latest_cached:
            logger.info(
                "ETag request failed with recoverable error, using latest cached "
                "version of %s: %s",
                url,
                latest_cached,
            )
            meta = Meta.from_path(_meta_file_path(latest_cached))
            return latest_cached, meta.etag
        else:
            logger.error(
                "ETag request failed with recoverable error, "
                "but no cached version of %s could be found",
                url,
            )
            raise

    filename = resource_to_filename(url, etag)

    # Get cache path to put the file.
    cache_path = cache_dir / filename

    # Multiple processes may be trying to cache the same file at once, so we need
    # to be a little careful to avoid race conditions. We do this using a lock file.
    # Only one process can own this lock file at a time, and a process will block
    # on the call to `lock.acquire()` until the process currently holding the lock
    # releases it.
    logger.debug("waiting to acquire lock on %s", cache_path)
    with FileLock(_lock_file_path(cache_path), read_only_ok=True):
        if os.path.exists(cache_path):
            logger.info("cache of %s is up-to-date", url)
        else:
            size = client.get_size()
            with CacheFile(cache_path) as cache_file:
                logger.info("%s not found in cache, downloading to %s", url, cache_path)

                from .progress import BufferedWriterWithProgress, get_download_progress

                start_and_cleanup = progress is None
                progress = progress or get_download_progress(quiet=quiet)

                if start_and_cleanup:
                    progress.start()

                try:
                    display_url = url if len(url) <= 50 else f"{url[:49]}\N{horizontal ellipsis}"
                    task_id = progress.add_task(f"Downloading [cyan i]{display_url}[/]", total=size)
                    writer_with_progress = BufferedWriterWithProgress(cache_file, progress, task_id)
                    client.get_resource(writer_with_progress)
                    progress.update(
                        task_id,
                        total=writer_with_progress.total_written,
                        completed=writer_with_progress.total_written,
                    )
                finally:
                    if start_and_cleanup:
                        progress.stop()

            logger.debug("creating metadata file for %s", cache_path)
            meta = Meta.new(
                url,
                cache_path,
                etag=etag,
            )
            meta.to_file()

    return cache_path, etag



import json
import logging
import os
from collections import defaultdict
from datetime import datetime
from typing import Any, DefaultDict, List, Mapping, Optional

from snorkel.types import Config


class LogWriterConfig(Config):
    """Manager for checkpointing model.

    Parameters
    ----------
    log_dir
        The root directory where logs should be saved
    run_name
        The name of this particular run (defaults to date-time combination if None)
    """

    log_dir: str = "logs"
    run_name: Optional[str] = None


class LogWriter:
    """A class for writing logs.

    Parameters
    ----------
    kwargs
        Settings to merge into LogWriterConfig

    Attributes
    ----------
    config
        Merged configuration
    run_name
        Name of run if provided, otherwise date-time combination
    log_dir
        The root directory where logs should be saved
    run_log
        Dictionary of scalar values to log, keyed by value name
    """

    def __init__(self, **kwargs: Any) -> None:
        self.config = LogWriterConfig(**kwargs)

        self.run_name = self.config.run_name
        if self.run_name is None:
            date = datetime.now().strftime("%Y_%m_%d")
            time = datetime.now().strftime("%H_%M_%S")
            self.run_name = f"{date}/{time}/"

        self.log_dir = os.path.join(self.config.log_dir, self.run_name)
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        self.run_log: DefaultDict[str, List[List[float]]] = defaultdict(list)

    def add_scalar(self, name: str, value: float, step: float) -> None:
        """Log a scalar variable.

        Parameters
        ----------
        name
            Name of the scalar collection
        value
            Value of scalar
        step
            Step axis value
        """
        # Note: storing as list for JSON roundtripping
        self.run_log[name].append([step, value])

    def write_config(
        self, config: Config, config_filename: str = "config.json"
    ) -> None:
        """Dump the config to file.

        Parameters
        ----------
        config
            JSON-compatible config to write to file
        config_filename
            Name of file in logging directory to write to
        """
        self.write_json(config._asdict(), config_filename)

    def write_log(self, log_filename: str) -> None:
        """Dump the scalar value log to file.

        Parameters
        ----------
        log_filename
            Name of file in logging directory to write to
        """
        self.write_json(self.run_log, log_filename)

    def write_text(self, text: str, filename: str) -> None:
        """Dump user-provided text to filename (e.g., the launch command).

        Parameters
        ----------
        text
            Text to write
        filename
            Name of file in logging directory to write to
        """
        text_path = os.path.join(self.log_dir, filename)
        with open(text_path, "w") as f:
            f.write(text)

    def write_json(self, dict_to_write: Mapping[str, Any], filename: str) -> None:
        """Dump a JSON-compatbile object to root log directory.

        Parameters
        ----------
        dict_to_write
            JSON-compatbile object to log
        filename
            Name of file in logging directory to write to
        """
        if not filename.endswith(".json"):  # pragma: no cover
            logging.warning(
                f"Using write_json() method with a filename without a .json extension: {filename}"
            )
        log_path = os.path.join(self.log_dir, filename)
        with open(log_path, "w") as f:
            json.dump(dict_to_write, f)

    def cleanup(self) -> None:
        """Perform final operations and close writer if necessary."""
        self.write_log("log.json")



from typing import Any, Callable, List, Mapping, Optional

from ...labeling.lf import LabelingFunction
from ...preprocess import BasePreprocessor


class SlicingFunction(LabelingFunction):
    """Base class for slicing functions.

    See ``snorkel.labeling.lf.LabelingFunction`` for details.
    """

    pass


class slicing_function:
    """Decorator to define a SlicingFunction object from a function.

    Parameters
    ----------
    name
        Name of the SF
    resources
        Slicing resources passed in to ``f`` via ``kwargs``
    preprocessors
        Preprocessors to run on data points before SF execution

    Examples
    --------
    >>> @slicing_function()
    ... def f(x):
    ...     return x.a > 42
    >>> f
    SlicingFunction f, Preprocessors: []
    >>> from types import SimpleNamespace
    >>> x = SimpleNamespace(a=90, b=12)
    >>> f(x)
    True

    >>> @slicing_function(name="my_sf")
    ... def g(x):
    ...     return 0 if x.a > 42 else -1
    >>> g
    SlicingFunction my_sf, Preprocessors: []
    """

    def __init__(
        self,
        name: Optional[str] = None,
        resources: Optional[Mapping[str, Any]] = None,
        pre: Optional[List[BasePreprocessor]] = None,
    ) -> None:
        if callable(name):
            raise ValueError("Looks like this decorator is missing parentheses!")
        self.name = name
        self.resources = resources
        self.pre = pre

    def __call__(self, f: Callable[..., int]) -> SlicingFunction:
        """Wrap a function to create a ``SlicingFunction``.

        Parameters
        ----------
        f
            Function that implements the core LF logic

        Returns
        -------
        SlicingFunction
            New ``SlicingFunction`` executing logic in wrapped function
        """
        name = self.name or f.__name__
        return SlicingFunction(name=name, f=f, resources=self.resources, pre=self.pre)



import inspect
import pickle
from collections.abc import Hashable
from types import SimpleNamespace
from typing import Any, Callable, Dict, List, Mapping, Optional

import numpy as np
import pandas as pd

from ..types import DataPoint, FieldMap, HashingFunction

MapFunction = Callable[[DataPoint], Optional[DataPoint]]


def get_parameters(
    f: Callable[..., Any], allow_args: bool = False, allow_kwargs: bool = False
) -> List[str]:
    """Get names of function parameters."""
    params = inspect.getfullargspec(f)
    if not allow_args and params[1] is not None:
        raise ValueError(f"Function {f.__name__} should not have *args")
    if not allow_kwargs and params[2] is not None:
        raise ValueError(f"Function {f.__name__} should not have **kwargs")
    return params[0]


def is_hashable(obj: Any) -> bool:
    """Test if object is hashable via duck typing.

    NB: not using ``collections.Hashable`` as some objects
    (e.g. pandas.Series) have a ``__hash__`` method to throw
    a more specific exception.
    """
    try:
        hash(obj)
        return True
    except Exception:
        return False


def get_hashable(obj: Any) -> Hashable:
    """Get a hashable version of a potentially unhashable object.

    This helper is used for caching mapper outputs of data points.
    For common data point formats (e.g. SimpleNamespace, pandas.Series),
    produces hashable representations of the values using a ``frozenset``.
    For objects like ``pandas.Series``, the name/index indentifier is dropped.

    Parameters
    ----------
    obj
        Object to get hashable version of

    Returns
    -------
    Hashable
        Hashable representation of object values

    Raises
    ------
    ValueError
        No hashable proxy for object
    """
    # If hashable already, just return
    if is_hashable(obj):
        return obj
    # Get dictionary from SimpleNamespace
    if isinstance(obj, SimpleNamespace):
        obj = vars(obj)
    # For dictionaries or pd.Series, construct a frozenset from items
    # Also recurse on values in case they aren't hashable
    if isinstance(obj, (dict, pd.Series)):
        return frozenset((k, get_hashable(v)) for k, v in obj.items())
    # For lists, recurse on values
    if isinstance(obj, (list, tuple)):
        return tuple(get_hashable(v) for v in obj)
    # For NumPy arrays, hash the byte representation of the data array
    if isinstance(obj, np.ndarray):
        return obj.data.tobytes()
    raise ValueError(f"Object {obj} has no hashing proxy.")


class BaseMapper:
    """Base class for ``Mapper`` and ``LambdaMapper``.

    Implements nesting, memoization, and deep copy functionality.
    Used primarily for type checking.

    Parameters
    ----------
    name
        Name of the mapper
    pre
        Mappers to run before this mapper is executed
    memoize
        Memoize mapper outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)

    Raises
    ------
    NotImplementedError
        Subclasses need to implement ``_generate_mapped_data_point``

    Attributes
    ----------
    memoize
        Memoize mapper outputs?
    """

    def __init__(
        self,
        name: str,
        pre: List["BaseMapper"],
        memoize: bool,
        memoize_key: Optional[HashingFunction] = None,
    ) -> None:
        if memoize_key is None:
            memoize_key = get_hashable
        self.name = name
        self._pre = pre
        self._memoize_key = memoize_key
        self.memoize = memoize
        self.reset_cache()

    def reset_cache(self) -> None:
        """Reset the memoization cache."""
        self._cache: Dict[DataPoint, DataPoint] = {}

    def _generate_mapped_data_point(self, x: DataPoint) -> Optional[DataPoint]:
        raise NotImplementedError

    def __call__(self, x: DataPoint) -> Optional[DataPoint]:
        """Run mapping function on input data point.

        Deep copies the data point first so as not to make
        accidental in-place changes. If ``memoize`` is set to
        ``True``, an internal cache is checked for results. If
        no cached results are found, the computed results are
        added to the cache.

        Parameters
        ----------
        x
            Data point to run mapping function on

        Returns
        -------
        DataPoint
            Mapped data point of same format but possibly different fields
        """
        if self.memoize:
            # NB: don't do ``self._cache.get(...)`` first in case cached value is ``None``
            x_hashable = self._memoize_key(x)
            if x_hashable in self._cache:
                return self._cache[x_hashable]
        # NB: using pickle roundtrip as a more robust deepcopy
        # As an example, calling deepcopy on a pd.Series or SimpleNamespace
        # with a dictionary attribute won't create a copy of the dictionary
        x_mapped = pickle.loads(pickle.dumps(x))
        for mapper in self._pre:
            x_mapped = mapper(x_mapped)
        x_mapped = self._generate_mapped_data_point(x_mapped)
        if self.memoize:
            self._cache[x_hashable] = x_mapped
        return x_mapped

    def __repr__(self) -> str:
        pre_str = f", Pre: {self._pre}"
        return f"{type(self).__name__} {self.name}{pre_str}"


class Mapper(BaseMapper):
    """Base class for any data point to data point mapping in the pipeline.

    Map data points to new data points by transforming, adding
    additional information, or decomposing into primitives. This module
    provides base classes for other operators like ``TransformationFunction``
    and ``Preprocessor``. We don't expect people to construct ``Mapper``
    objects directly.

    A Mapper maps an data point to a new data point, possibly with
    a different schema. Subclasses of Mapper need to implement the
    ``run`` method, which takes fields of the data point as input
    and outputs new fields for the mapped data point as a dictionary.
    The ``run`` method should only be called internally by the ``Mapper``
    object, not directly by a user.

    Mapper derivatives work for data points that have mutable attributes,
    like ``SimpleNamespace``, ``pd.Series``, or ``dask.Series``. An example
    of a data point type without mutable fields is ``pyspark.sql.Row``.
    Use ``snorkel.map.spark.make_spark_mapper`` for PySpark compatibility.

    For an example of a Mapper, see
        ``snorkel.preprocess.nlp.SpacyPreprocessor``

    Parameters
    ----------
    name
        Name of mapper
    field_names
        A map from attribute names of the incoming data points
        to the input argument names of the ``run`` method. If None,
        the parameter names in the function signature are used.
    mapped_field_names
        A map from output keys of the ``run`` method to attribute
        names of the output data points. If None, the original
        output keys are used.
    pre
        Mappers to run before this mapper is executed
    memoize
        Memoize mapper outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)

    Raises
    ------
    NotImplementedError
        Subclasses must implement the ``run`` method

    Attributes
    ----------
    field_names
        See above
    mapped_field_names
        See above
    memoize
        Memoize mapper outputs?
    """

    def __init__(
        self,
        name: str,
        field_names: Optional[Mapping[str, str]] = None,
        mapped_field_names: Optional[Mapping[str, str]] = None,
        pre: Optional[List[BaseMapper]] = None,
        memoize: bool = False,
        memoize_key: Optional[HashingFunction] = None,
    ) -> None:
        if field_names is None:
            # Parse field names from ``run(...)`` if not provided
            field_names = {k: k for k in get_parameters(self.run)[1:]}
        self.field_names = field_names
        self.mapped_field_names = mapped_field_names
        super().__init__(name, pre or [], memoize, memoize_key)

    def run(self, **kwargs: Any) -> Optional[FieldMap]:
        """Run the mapping operation using the input fields.

        The inputs to this function are fed by extracting the fields of
        the input data point using the keys of ``field_names``. The output field
        names are converted using ``mapped_field_names`` and added to the
        data point.

        Returns
        -------
        Optional[FieldMap]
            A mapping from canonical output field names to their values.

        Raises
        ------
        NotImplementedError
            Subclasses must implement this method
        """
        raise NotImplementedError

    def _update_fields(self, x: DataPoint, mapped_fields: FieldMap) -> DataPoint:
        # ``SimpleNamespace``, ``pd.Series``, and ``dask.Series`` objects all
        # have attribute setting.
        for k, v in mapped_fields.items():
            setattr(x, k, v)
        return x

    def _generate_mapped_data_point(self, x: DataPoint) -> Optional[DataPoint]:
        field_map = {k: getattr(x, v) for k, v in self.field_names.items()}
        mapped_fields = self.run(**field_map)
        if mapped_fields is None:
            return None
        if self.mapped_field_names is not None:
            mapped_fields = {
                v: mapped_fields[k] for k, v in self.mapped_field_names.items()
            }
        return self._update_fields(x, mapped_fields)


class LambdaMapper(BaseMapper):
    """Define a mapper from a function.

    Convenience class for mappers that execute a simple
    function with no set up. The function should map from
    an input data point to a new data point directly, unlike
    ``Mapper.run``. The original data point will not be updated,
    so in-place operations are safe.

    Parameters
    ----------
    name
        Name of mapper
    f
        Function executing the mapping operation
    pre
        Mappers to run before this mapper is executed
    memoize
        Memoize mapper outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)
    """

    def __init__(
        self,
        name: str,
        f: MapFunction,
        pre: Optional[List[BaseMapper]] = None,
        memoize: bool = False,
        memoize_key: Optional[HashingFunction] = None,
    ) -> None:
        self._f = f
        super().__init__(name, pre or [], memoize, memoize_key)

    def _generate_mapped_data_point(self, x: DataPoint) -> Optional[DataPoint]:
        return self._f(x)


class lambda_mapper:
    """Decorate a function to define a LambdaMapper object.

    Example
    -------
    >>> @lambda_mapper()
    ... def concatenate_text(x):
    ...     x.article = f"{x.title} {x.body}"
    ...     return x
    >>> isinstance(concatenate_text, LambdaMapper)
    True
    >>> from types import SimpleNamespace
    >>> x = SimpleNamespace(title="my title", body="my text")
    >>> concatenate_text(x).article
    'my title my text'

    Parameters
    ----------
    name
        Name of mapper. If None, uses the name of the wrapped function.
    pre
        Mappers to run before this mapper is executed
    memoize
        Memoize mapper outputs?
    memoize_key
        Hashing function to handle the memoization (default to snorkel.map.core.get_hashable)

    Attributes
    ----------
    memoize
        Memoize mapper outputs?
    """

    def __init__(
        self,
        name: Optional[str] = None,
        pre: Optional[List[BaseMapper]] = None,
        memoize: bool = False,
        memoize_key: Optional[HashingFunction] = None,
    ) -> None:
        if callable(name):
            raise ValueError("Looks like this decorator is missing parentheses!")
        self.name = name
        self.pre = pre
        self.memoize = memoize
        self.memoize_key = memoize_key

    def __call__(self, f: MapFunction) -> LambdaMapper:
        """Wrap a function to create a ``LambdaMapper``.

        Parameters
        ----------
        f
            Function executing the mapping operation

        Returns
        -------
        LambdaMapper
            New ``LambdaMapper`` executing operation in wrapped function
        """
        name = self.name or f.__name__
        return LambdaMapper(
            name=name,
            f=f,
            pre=self.pre,
            memoize=self.memoize,
            memoize_key=self.memoize_key,
        )



import logging
import os
import subprocess
import sys
import importlib
from pathlib import Path
from ekorpkit.base import _is_dir, _is_file


log = logging.getLogger(__name__)


def gitclone(url, targetdir=None, verbose=False):
    if targetdir:
        res = subprocess.run(
            ["git", "clone", url, targetdir], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
    else:
        res = subprocess.run(
            ["git", "clone", url], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        log.info(res)


def pip(
    name,
    upgrade=False,
    prelease=False,
    editable=False,
    quiet=True,
    find_links=None,
    requirement=None,
    force_reinstall=False,
    verbose=False,
    **kwargs,
):
    _cmd = ["pip", "install"]
    if upgrade:
        _cmd.append("--upgrade")
    if prelease:
        _cmd.append("--pre")
    if editable:
        _cmd.append("--editable")
    if quiet:
        _cmd.append("--quiet")
    if find_links:
        _cmd += ["--find-links", find_links]
    if requirement:
        _cmd.append("--requirement")
    if force_reinstall:
        _cmd.append("--force-reinstall")
    for k in kwargs:
        k = k.replace("_", "-")
        _cmd.append(f"--{k}")
    _cmd.append(name)
    if verbose:
        log.info(f"Installing: {' '.join(_cmd)}")
    res = subprocess.run(_cmd, stdout=subprocess.PIPE).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        log.info(res)


def pipi(name, verbose=False):
    res = subprocess.run(
        ["pip", "install", name], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        log.info(res)


def pipie(name, verbose=False):
    res = subprocess.run(
        ["git", "install", "-e", name], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        log.info(res)


def apti(name, verbose=False):
    res = subprocess.run(
        ["apt", "install", name], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        log.info(res)


def load_module_from_file(name, libpath, specname=None):
    module_path = os.path.join(libpath, name.replace(".", os.path.sep))
    if _is_file(module_path + ".py"):
        module_path = module_path + ".py"
    else:
        if _is_dir(module_path):
            module_path = os.path.join(module_path, "__init__.py")
        else:
            module_path = str(Path(module_path).parent / "__init__.py")

    spec = importlib.util.spec_from_file_location(name, module_path)
    module = importlib.util.module_from_spec(spec)
    if not specname:
        specname = spec.name
    sys.modules[specname] = module
    spec.loader.exec_module(module)


def ensure_import_module(name, libpath, liburi, specname=None, syspath=None):
    try:
        if specname:
            importlib.import_module(specname)
        else:
            importlib.import_module(name)
        log.info(f"{name} imported")
    except:
        if not os.path.exists(libpath):
            log.info(f"{libpath} not found, cloning from {liburi}")
            gitclone(liburi, libpath)
        if not syspath:
            syspath = libpath
        if syspath not in sys.path:
            sys.path.append(syspath)
        load_module_from_file(name, syspath, specname)
        specname = specname or name
        log.info(f"{name} not imported, loading from {syspath} as {specname}")
